# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Contains the definition of the Inception V4 architecture.

As described in http://arxiv.org/abs/1602.07261.

  Inception-v4, Inception-ResNet and the Impact of Residual Connections
    on Learning
  Christian Szegedy, Sergey Ioffe, Vincent Vanhoucke, Alex Alemi
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v1 as tf
import tf_slim as slim

from nets import inception_utils


def block_inception_a(inputs, scope=None, reuse=None):
  """Builds Inception-A block for Inception v4 network."""
  # By default use stride=1 and SAME padding
  with slim.arg_scope([slim.conv2d, slim.avg_pool2d, slim.max_pool2d],
                      stride=1, padding='SAME'):
    with tf.variable_scope(
        scope, 'BlockInceptionA', [inputs], reuse=reuse):
      with tf.variable_scope('Branch_0'):
        branch_0 = slim.conv2d(inputs, 96, [1, 1], scope='Conv2d_0a_1x1')
      with tf.variable_scope('Branch_1'):
        branch_1 = slim.conv2d(inputs, 64, [1, 1], scope='Conv2d_0a_1x1')
        branch_1 = slim.conv2d(branch_1, 96, [3, 3], scope='Conv2d_0b_3x3')
      with tf.variable_scope('Branch_2'):
        branch_2 = slim.conv2d(inputs, 64, [1, 1], scope='Conv2d_0a_1x1')
        branch_2 = slim.conv2d(branch_2, 96, [3, 3], scope='Conv2d_0b_3x3')
        branch_2 = slim.conv2d(branch_2, 96, [3, 3], scope='Conv2d_0c_3x3')
      with tf.variable_scope('Branch_3'):
        branch_3 = slim.avg_pool2d(inputs, [3, 3], scope='AvgPool_0a_3x3')
        branch_3 = slim.conv2d(branch_3, 96, [1, 1], scope='Conv2d_0b_1x1')
      return tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])


def block_reduction_a(inputs, scope=None, reuse=None):
  """Builds Reduction-A block for Inception v4 network."""
  # By default use stride=1 and SAME padding
  with slim.arg_scope([slim.conv2d, slim.avg_pool2d, slim.max_pool2d],
                      stride=1, padding='SAME'):
    with tf.variable_scope(
        scope, 'BlockReductionA', [inputs], reuse=reuse):
      with tf.variable_scope('Branch_0'):
        branch_0 = slim.conv2d(inputs, 384, [3, 3], stride=2, padding='VALID',
                               scope='Conv2d_1a_3x3')
      with tf.variable_scope('Branch_1'):
        branch_1 = slim.conv2d(inputs, 192, [1, 1], scope='Conv2d_0a_1x1')
        branch_1 = slim.conv2d(branch_1, 224, [3, 3], scope='Conv2d_0b_3x3')
        branch_1 = slim.conv2d(branch_1, 256, [3, 3], stride=2,
                               padding='VALID', scope='Conv2d_1a_3x3')
      with tf.variable_scope('Branch_2'):
        branch_2 = slim.max_pool2d(inputs, [3, 3], stride=2, padding='VALID',
                                   scope='MaxPool_1a_3x3')
      return tf.concat(axis=3, values=[branch_0, branch_1, branch_2])


def block_inception_b(inputs, scope=None, reuse=None):
  """Builds Inception-B block for Inception v4 network."""
  # By default use stride=1 and SAME padding
  with slim.arg_scope([slim.conv2d, slim.avg_pool2d, slim.max_pool2d],
                      stride=1, padding='SAME'):
    with tf.variable_scope(
        scope, 'BlockInceptionB', [inputs], reuse=reuse):
      with tf.variable_scope('Branch_0'):
        branch_0 = slim.conv2d(inputs, 384, [1, 1], scope='Conv2d_0a_1x1')
      with tf.variable_scope('Branch_1'):
        branch_1 = slim.conv2d(inputs, 192, [1, 1], scope='Conv2d_0a_1x1')
        branch_1 = slim.conv2d(branch_1, 224, [1, 7], scope='Conv2d_0b_1x7')
        branch_1 = slim.conv2d(branch_1, 256, [7, 1], scope='Conv2d_0c_7x1')
      with tf.variable_scope('Branch_2'):
        branch_2 = slim.conv2d(inputs, 192, [1, 1], scope='Conv2d_0a_1x1')
        branch_2 = slim.conv2d(branch_2, 192, [7, 1], scope='Conv2d_0b_7x1')
        branch_2 = slim.conv2d(branch_2, 224, [1, 7], scope='Conv2d_0c_1x7')
        branch_2 = slim.conv2d(branch_2, 224, [7, 1], scope='Conv2d_0d_7x1')
        branch_2 = slim.conv2d(branch_2, 256, [1, 7], scope='Conv2d_0e_1x7')
      with tf.variable_scope('Branch_3'):
        branch_3 = slim.avg_pool2d(inputs, [3, 3], scope='AvgPool_0a_3x3')
        branch_3 = slim.conv2d(branch_3, 128, [1, 1], scope='Conv2d_0b_1x1')
      return tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])


def block_reduction_b(inputs, scope=None, reuse=None):
  """Builds Reduction-B block for Inception v4 network."""
  # By default use stride=1 and SAME padding
  with slim.arg_scope([slim.conv2d, slim.avg_pool2d, slim.max_pool2d],
                      stride=1, padding='SAME'):
    with tf.variable_scope(
        scope, 'BlockReductionB', [inputs], reuse=reuse):
      with tf.variable_scope('Branch_0'):
        branch_0 = slim.conv2d(inputs, 192, [1, 1], scope='Conv2d_0a_1x1')
        branch_0 = slim.conv2d(branch_0, 192, [3, 3], stride=2,
                               padding='VALID', scope='Conv2d_1a_3x3')
      with tf.variable_scope('Branch_1'):
        branch_1 = slim.conv2d(inputs, 256, [1, 1], scope='Conv2d_0a_1x1')
        branch_1 = slim.conv2d(branch_1, 256, [1, 7], scope='Conv2d_0b_1x7')
        branch_1 = slim.conv2d(branch_1, 320, [7, 1], scope='Conv2d_0c_7x1')
        branch_1 = slim.conv2d(branch_1, 320, [3, 3], stride=2,
                               padding='VALID', scope='Conv2d_1a_3x3')
      with tf.variable_scope('Branch_2'):
        branch_2 = slim.max_pool2d(inputs, [3, 3], stride=2, padding='VALID',
                                   scope='MaxPool_1a_3x3')
      return tf.concat(axis=3, values=[branch_0, branch_1, branch_2])


def block_inception_c(inputs, scope=None, reuse=None):
  """Builds Inception-C block for Inception v4 network."""
  # By default use stride=1 and SAME padding
  with slim.arg_scope([slim.conv2d, slim.avg_pool2d, slim.max_pool2d],
                      stride=1, padding='SAME'):
    with tf.variable_scope(
        scope, 'BlockInceptionC', [inputs], reuse=reuse):
      with tf.variable_scope('Branch_0'):
        branch_0 = slim.conv2d(inputs, 256, [1, 1], scope='Conv2d_0a_1x1')
      with tf.variable_scope('Branch_1'):
        branch_1 = slim.conv2d(inputs, 384, [1, 1], scope='Conv2d_0a_1x1')
        branch_1 = tf.concat(axis=3, values=[
            slim.conv2d(branch_1, 256, [1, 3], scope='Conv2d_0b_1x3'),
            slim.conv2d(branch_1, 256, [3, 1], scope='Conv2d_0c_3x1')])
      with tf.variable_scope('Branch_2'):
        branch_2 = slim.conv2d(inputs, 384, [1, 1], scope='Conv2d_0a_1x1')
        branch_2 = slim.conv2d(branch_2, 448, [3, 1], scope='Conv2d_0b_3x1')
        branch_2 = slim.conv2d(branch_2, 512, [1, 3], scope='Conv2d_0c_1x3')
        branch_2 = tf.concat(axis=3, values=[
            slim.conv2d(branch_2, 256, [1, 3], scope='Conv2d_0d_1x3'),
            slim.conv2d(branch_2, 256, [3, 1], scope='Conv2d_0e_3x1')])
      with tf.variable_scope('Branch_3'):
        branch_3 = slim.avg_pool2d(inputs, [3, 3], scope='AvgPool_0a_3x3')
        branch_3 = slim.conv2d(branch_3, 256, [1, 1], scope='Conv2d_0b_1x1')
      return tf.concat(axis=3, values=[branch_0, branch_1, branch_2, branch_3])


def inception_v4_base(inputs, final_endpoint='Mixed_7d', scope=None):
  """Creates the Inception V4 network up to the given final endpoint.

  Args:
    inputs: a 4-D tensor of size [batch_size, height, width, 3].
    final_endpoint: specifies the endpoint to construct the network up to.
      It can be one of [ 'Conv2d_1a_3x3', 'Conv2d_2a_3x3', 'Conv2d_2b_3x3',
      'Mixed_3a', 'Mixed_4a', 'Mixed_5a', 'Mixed_5b', 'Mixed_5c', 'Mixed_5d',
      'Mixed_5e', 'Mixed_6a', 'Mixed_6b', 'Mixed_6c', 'Mixed_6d', 'Mixed_6e',
      'Mixed_6f', 'Mixed_6g', 'Mixed_6h', 'Mixed_7a', 'Mixed_7b', 'Mixed_7c',
      'Mixed_7d']
    scope: Optional variable_scope.

  Returns:
    logits: the logits outputs of the model.
    end_points: the set of end_points from the inception model.

  Raises:
    ValueError: if final_endpoint is not set to one of the predefined values,
  """
  end_points = {}

  def add_and_check_final(name, net):
    end_points[name] = net
    return name == final_endpoint

  with tf.variable_scope(scope, 'InceptionV4', [inputs]):
    with slim.arg_scope([slim.conv2d, slim.max_pool2d, slim.avg_pool2d],
                        stride=1, padding='SAME'):
      # 299 x 299 x 3
      net = slim.conv2d(inputs, 32, [3, 3], stride=2,
                        padding='VALID', scope='Conv2d_1a_3x3')
      if add_and_check_final('Conv2d_1a_3x3', net): return net, end_points
      # 149 x 149 x 32
      net = slim.conv2d(net, 32, [3, 3], padding='VALID',
                        scope='Conv2d_2a_3x3')
      if add_and_check_final('Conv2d_2a_3x3', net): return net, end_points
      # 147 x 147 x 32
      net = slim.conv2d(net, 64, [3, 3], scope='Conv2d_2b_3x3')
      if add_and_check_final('Conv2d_2b_3x3', net): return net, end_points
      # 147 x 147 x 64
      with tf.variable_scope('Mixed_3a'):
        with tf.variable_scope('Branch_0'):
          branch_0 = slim.max_pool2d(net, [3, 3], stride=2, padding='VALID',
                                     scope='MaxPool_0a_3x3')
        with tf.variable_scope('Branch_1'):
          branch_1 = slim.conv2d(net, 96, [3, 3], stride=2, padding='VALID',
                                 scope='Conv2d_0a_3x3')
        net = tf.concat(axis=3, values=[branch_0, branch_1])
        if add_and_check_final('Mixed_3a', net): return net, end_points

      # 73 x 73 x 160
      with tf.variable_scope('Mixed_4a'):
        with tf.variable_scope('Branch_0'):
          branch_0 = slim.conv2d(net, 64, [1, 1], scope='Conv2d_0a_1x1')
          branch_0 = slim.conv2d(branch_0, 96, [3, 3], padding='VALID',
                                 scope='Conv2d_1a_3x3')
        with tf.variable_scope('Branch_1'):
          branch_1 = slim.conv2d(net, 64, [1, 1], scope='Conv2d_0a_1x1')
          branch_1 = slim.conv2d(branch_1, 64, [1, 7], scope='Conv2d_0b_1x7')
          branch_1 = slim.conv2d(branch_1, 64, [7, 1], scope='Conv2d_0c_7x1')
          branch_1 = slim.conv2d(branch_1, 96, [3, 3], padding='VALID',
                                 scope='Conv2d_1a_3x3')
        net = tf.concat(axis=3, values=[branch_0, branch_1])
        if add_and_check_final('Mixed_4a', net): return net, end_points

      # 71 x 71 x 192
      with tf.variable_scope('Mixed_5a'):
        with tf.variable_scope('Branch_0'):
          branch_0 = slim.conv2d(net, 192, [3, 3], stride=2, padding='VALID',
                                 scope='Conv2d_1a_3x3')
        with tf.variable_scope('Branch_1'):
          branch_1 = slim.max_pool2d(net, [3, 3], stride=2, padding='VALID',
                                     scope='MaxPool_1a_3x3')
        net = tf.concat(axis=3, values=[branch_0, branch_1])
        if add_and_check_final('Mixed_5a', net): return net, end_points

      # 35 x 35 x 384
      # 4 x Inception-A blocks
      for idx in range(4):
        block_scope = 'Mixed_5' + chr(ord('b') + idx)
        net = block_inception_a(net, block_scope)
        if add_and_check_final(block_scope, net): return net, end_points

      # 35 x 35 x 384
      # Reduction-A block
      net = block_reduction_a(net, 'Mixed_6a')
      if add_and_check_final('Mixed_6a', net): return net, end_points

      # 17 x 17 x 1024
      # 7 x Inception-B blocks
      for idx in range(7):
        block_scope = 'Mixed_6' + chr(ord('b') + idx)
        net = block_inception_b(net, block_scope)
        if add_and_check_final(block_scope, net): return net, end_points

      # 17 x 17 x 1024
      # Reduction-B block
      net = block_reduction_b(net, 'Mixed_7a')
      if add_and_check_final('Mixed_7a', net): return net, end_points

      # 8 x 8 x 1536
      # 3 x Inception-C blocks
      for idx in range(3):
        block_scope = 'Mixed_7' + chr(ord('b') + idx)
        net = block_inception_c(net, block_scope)
        if add_and_check_final(block_scope, net): return net, end_points
  raise ValueError('Unknown final endpoint %s' % final_endpoint)


def inception_v4(inputs, num_classes=1001, is_training=True,
                 dropout_keep_prob=0.8,
                 reuse=None,
                 scope='InceptionV4',
                 create_aux_logits=True):
  """Creates the Inception V4 model.

  Args:
    inputs: a 4-D tensor of size [batch_size, height, width, 3].
    num_classes: number of predicted classes. If 0 or None, the logits layer
      is omitted and the input features to the logits layer (before dropout)
      are returned instead.
    is_training: whether is training or not.
    dropout_keep_prob: float, the fraction to keep before final layer.
    reuse: whether or not the network and its variables should be reused. To be
      able to reuse 'scope' must be given.
    scope: Optional variable_scope.
    create_aux_logits: Whether to include the auxiliary logits.

  Returns:
    net: a Tensor with the logits (pre-softmax activations) if num_classes
      is a non-zero integer, or the non-dropped input to the logits layer
      if num_classes is 0 or None.
    end_points: the set of end_points from the inception model.
  """
  end_points = {}
  with tf.variable_scope(
      scope, 'InceptionV4', [inputs], reuse=reuse) as scope:
    with slim.arg_scope([slim.batch_norm, slim.dropout],
                        is_training=is_training):
      net, end_points = inception_v4_base(inputs, scope=scope)

      with slim.arg_scope([slim.conv2d, slim.max_pool2d, slim.avg_pool2d],
                          stride=1, padding='SAME'):
        # Auxiliary Head logits
        if create_aux_logits and num_classes:
          with tf.variable_scope('AuxLogits'):
            # 17 x 17 x 1024
            aux_logits = end_points['Mixed_6h']
            aux_logits = slim.avg_pool2d(aux_logits, [5, 5], stride=3,
                                         padding='VALID',
                                         scope='AvgPool_1a_5x5')
            aux_logits = slim.conv2d(aux_logits, 128, [1, 1],
                                     scope='Conv2d_1b_1x1')
            aux_logits = slim.conv2d(aux_logits, 768,
                                     aux_logits.get_shape()[1:3],
                                     padding='VALID', scope='Conv2d_2a')
            aux_logits = slim.flatten(aux_logits)
            aux_logits = slim.fully_connected(aux_logits, num_classes,
                                              activation_fn=None,
                                              scope='Aux_logits')
            end_points['AuxLogits'] = aux_logits

        # Final pooling and prediction
        # TODO(sguada,arnoegw): Consider adding a parameter global_pool which
        # can be set to False to disable pooling here (as in resnet_*()).
        with tf.variable_scope('Logits'):
          # 8 x 8 x 1536
          kernel_size = net.get_shape()[1:3]
          if kernel_size.is_fully_defined():
            net = slim.avg_pool2d(net, kernel_size, padding='VALID',
                                  scope='AvgPool_1a')
          else:
            net = tf.reduce_mean(
                input_tensor=net,
                axis=[1, 2],
                keepdims=True,
                name='global_pool')
          end_points['global_pool'] = net
          if not num_classes:
            return net, end_points
          # 1 x 1 x 1536
          net = slim.dropout(net, dropout_keep_prob, scope='Dropout_1b')
          net = slim.flatten(net, scope='PreLogitsFlatten')
          end_points['PreLogitsFlatten'] = net
          # 1536
          logits = slim.fully_connected(net, num_classes, activation_fn=None,
                                        scope='Logits')
          end_points['Logits'] = logits
          end_points['Predictions'] = tf.nn.softmax(logits, name='Predictions')
    return logits, end_points
inception_v4.default_image_size = 299


inception_v4_arg_scope = inception_utils.inception_arg_scope



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for google3.third_party.tensorflow_models.slim.nets.mobilenet.mobilenet_v3."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v1 as tf

from nets.mobilenet import mobilenet_v3
from google3.testing.pybase import parameterized


class MobilenetV3Test(tf.test.TestCase, parameterized.TestCase):

  # pylint: disable = g-unreachable-test-method
  def assertVariablesHaveNormalizerFn(self, use_groupnorm):
    global_variables = [v.name for v in tf.global_variables()]
    has_batch_norm = False
    has_group_norm = False
    for global_variable in global_variables:
      if 'BatchNorm' in global_variable:
        has_batch_norm = True
      if 'GroupNorm' in global_variable:
        has_group_norm = True
    if use_groupnorm:
      self.assertFalse(has_batch_norm)
      self.assertTrue(has_group_norm)
    else:
      self.assertTrue(has_batch_norm)
      self.assertFalse(has_group_norm)

  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetV3Large(self, use_groupnorm):
    logits, endpoints = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        use_groupnorm=use_groupnorm)
    self.assertEqual(endpoints['layer_19'].shape, [1, 1, 1, 1280])
    self.assertEqual(logits.shape, [1, 1001])
    self.assertVariablesHaveNormalizerFn(use_groupnorm)

  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetV3Small(self, use_groupnorm):
    _, endpoints = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        conv_defs=mobilenet_v3.V3_SMALL,
        use_groupnorm=use_groupnorm)
    self.assertEqual(endpoints['layer_15'].shape, [1, 1, 1, 1024])
    self.assertVariablesHaveNormalizerFn(use_groupnorm)

  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetEdgeTpu(self, use_groupnorm):
    _, endpoints = mobilenet_v3.edge_tpu(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        use_groupnorm=use_groupnorm)
    self.assertIn('Inference mode is created by default',
                  mobilenet_v3.edge_tpu.__doc__)
    self.assertEqual(endpoints['layer_24'].shape, [1, 7, 7, 1280])
    self.assertStartsWith(
        endpoints['layer_24'].name, 'MobilenetEdgeTPU')
    self.assertVariablesHaveNormalizerFn(use_groupnorm)

  def testMobilenetEdgeTpuChangeScope(self):
    _, endpoints = mobilenet_v3.edge_tpu(
        tf.placeholder(tf.float32, (1, 224, 224, 3)), scope='Scope')
    self.assertStartsWith(
        endpoints['layer_24'].name, 'Scope')

  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetV3BaseOnly(self, use_groupnorm):
    result, endpoints = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        conv_defs=mobilenet_v3.V3_LARGE,
        use_groupnorm=use_groupnorm,
        base_only=True,
        final_endpoint='layer_17')
    # Get the latest layer before average pool.
    self.assertEqual(endpoints['layer_17'].shape, [1, 7, 7, 960])
    self.assertEqual(result, endpoints['layer_17'])
    self.assertVariablesHaveNormalizerFn(use_groupnorm)

  def testMobilenetV3BaseOnly_VariableInput(self):
    result, endpoints = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (None, None, None, 3)),
        conv_defs=mobilenet_v3.V3_LARGE,
        base_only=True,
        final_endpoint='layer_17')
    # Get the latest layer before average pool.
    self.assertEqual(endpoints['layer_17'].shape.as_list(),
                     [None, None, None, 960])
    self.assertEqual(result, endpoints['layer_17'])

  # Use reduce mean for pooling and check for operation 'ReduceMean' in graph
  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetV3WithReduceMean(self, use_groupnorm):
    _, _ = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        conv_defs=mobilenet_v3.V3_SMALL,
        use_groupnorm=use_groupnorm,
        use_reduce_mean_for_pooling=True)
    g = tf.get_default_graph()
    reduce_mean = [v for v in g.get_operations() if 'ReduceMean' in v.name]
    self.assertNotEmpty(reduce_mean)
    self.assertVariablesHaveNormalizerFn(use_groupnorm)

  @parameterized.named_parameters(('without_groupnorm', False),
                                  ('with_groupnorm', True))
  def testMobilenetV3WithOutReduceMean(self, use_groupnorm):
    _, _ = mobilenet_v3.mobilenet(
        tf.placeholder(tf.float32, (1, 224, 224, 3)),
        conv_defs=mobilenet_v3.V3_SMALL,
        use_groupnorm=use_groupnorm,
        use_reduce_mean_for_pooling=False)
    g = tf.get_default_graph()
    reduce_mean = [v for v in g.get_operations() if 'ReduceMean' in v.name]
    self.assertEmpty(reduce_mean)
    self.assertVariablesHaveNormalizerFn(use_groupnorm)


if __name__ == '__main__':
  # absltest.main()
  tf.test.main()



"""
Usage:
  # From tensorflow/models/
  # Create train data:
  python generate_tfrecord.py --csv_input=data/train_labels.csv  --output_path=train.record

  # Create test data:
  python generate_tfrecord.py --csv_input=data/test_labels.csv  --output_path=test.record
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

import os
import io
import pandas as pd
import tensorflow as tf

from PIL import Image
from object_detection.utils import dataset_util
from collections import namedtuple, OrderedDict

flags = tf.app.flags
flags.DEFINE_string('csv_input', '', 'Path to the CSV input')
flags.DEFINE_string('output_path', '', 'Path to output TFRecord')
FLAGS = flags.FLAGS


# TO-DO replace this with label map
def class_text_to_int(row_label):
    if row_label == 'OpticDisk':
        return 1
    elif row_label == 'Macula':
        return 2
    else:
        None


def split(df, group):
    data = namedtuple('data', ['filename', 'object'])
    gb = df.groupby(group)
    return [data(filename, gb.get_group(x)) for filename, x in zip(gb.groups.keys(), gb.groups)]


def create_tf_example(group, path):
    with tf.gfile.GFile(os.path.join(path, '{}'.format(group.filename)), 'rb') as fid:
        encoded_jpg = fid.read()
    encoded_jpg_io = io.BytesIO(encoded_jpg)
    image = Image.open(encoded_jpg_io)
    width, height = image.size

    filename = group.filename.encode('utf8')
    image_format = b'jpg'
    xmins = []
    xmaxs = []
    ymins = []
    ymaxs = []
    classes_text = []
    classes = []

    for index, row in group.object.iterrows():
        xmins.append(row['xmin'] / width)
        xmaxs.append(row['xmax'] / width)
        ymins.append(row['ymin'] / height)
        ymaxs.append(row['ymax'] / height)
        classes_text.append(row['class'].encode('utf8'))
        classes.append(class_text_to_int(row['class']))

    tf_example = tf.train.Example(features=tf.train.Features(feature={
        'image/height': dataset_util.int64_feature(height),
        'image/width': dataset_util.int64_feature(width),
        'image/filename': dataset_util.bytes_feature(filename),
        'image/source_id': dataset_util.bytes_feature(filename),
        'image/encoded': dataset_util.bytes_feature(encoded_jpg),
        'image/format': dataset_util.bytes_feature(image_format),
        'image/object/bbox/xmin': dataset_util.float_list_feature(xmins),
        'image/object/bbox/xmax': dataset_util.float_list_feature(xmaxs),
        'image/object/bbox/ymin': dataset_util.float_list_feature(ymins),
        'image/object/bbox/ymax': dataset_util.float_list_feature(ymaxs),
        'image/object/class/text': dataset_util.bytes_list_feature(classes_text),
        'image/object/class/label': dataset_util.int64_list_feature(classes),
    }))
    return tf_example


def main(_):
    writer = tf.python_io.TFRecordWriter(FLAGS.output_path)
    path = os.path.join(os.getcwd(), 'images')
    examples = pd.read_csv(FLAGS.csv_input)
    grouped = split(examples, 'filename')
    for group in grouped:
        tf_example = create_tf_example(group, path)
        writer.write(tf_example.SerializeToString())

    writer.close()
    output_path = os.path.join(os.getcwd(), FLAGS.output_path)
    print('Successfully created the TFRecords: {}'.format(output_path))


if __name__ == '__main__':
    tf.app.run()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for post_processing_builder."""

import tensorflow.compat.v1 as tf
from google.protobuf import text_format
from ..builders import post_processing_builder
from ..protos import post_processing_pb2
from ..utils import test_case


class PostProcessingBuilderTest(test_case.TestCase):

  def test_build_non_max_suppressor_with_correct_parameters(self):
    post_processing_text_proto = """
      batch_non_max_suppression {
        score_threshold: 0.7
        iou_threshold: 0.6
        max_detections_per_class: 100
        max_total_detections: 300
        soft_nms_sigma: 0.4
      }
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    non_max_suppressor, _ = post_processing_builder.build(
        post_processing_config)
    self.assertEqual(non_max_suppressor.keywords['max_size_per_class'], 100)
    self.assertEqual(non_max_suppressor.keywords['max_total_size'], 300)
    self.assertAlmostEqual(non_max_suppressor.keywords['score_thresh'], 0.7)
    self.assertAlmostEqual(non_max_suppressor.keywords['iou_thresh'], 0.6)
    self.assertAlmostEqual(non_max_suppressor.keywords['soft_nms_sigma'], 0.4)

  def test_build_non_max_suppressor_with_correct_parameters_classagnostic_nms(
      self):
    post_processing_text_proto = """
      batch_non_max_suppression {
        score_threshold: 0.7
        iou_threshold: 0.6
        max_detections_per_class: 10
        max_total_detections: 300
        use_class_agnostic_nms: True
        max_classes_per_detection: 1
      }
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    non_max_suppressor, _ = post_processing_builder.build(
        post_processing_config)
    self.assertEqual(non_max_suppressor.keywords['max_size_per_class'], 10)
    self.assertEqual(non_max_suppressor.keywords['max_total_size'], 300)
    self.assertEqual(non_max_suppressor.keywords['max_classes_per_detection'],
                     1)
    self.assertEqual(non_max_suppressor.keywords['use_class_agnostic_nms'],
                     True)
    self.assertAlmostEqual(non_max_suppressor.keywords['score_thresh'], 0.7)
    self.assertAlmostEqual(non_max_suppressor.keywords['iou_thresh'], 0.6)

  def test_build_identity_score_converter(self):
    post_processing_text_proto = """
      score_converter: IDENTITY
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, score_converter = post_processing_builder.build(
        post_processing_config)
    self.assertEqual(score_converter.__name__, 'identity_with_logit_scale')
    def graph_fn():
      inputs = tf.constant([1, 1], tf.float32)
      outputs = score_converter(inputs)
      return outputs
    converted_scores = self.execute_cpu(graph_fn, [])
    self.assertAllClose(converted_scores, [1, 1])

  def test_build_identity_score_converter_with_logit_scale(self):
    post_processing_text_proto = """
      score_converter: IDENTITY
      logit_scale: 2.0
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, score_converter = post_processing_builder.build(post_processing_config)
    self.assertEqual(score_converter.__name__, 'identity_with_logit_scale')

    def graph_fn():
      inputs = tf.constant([1, 1], tf.float32)
      outputs = score_converter(inputs)
      return outputs
    converted_scores = self.execute_cpu(graph_fn, [])
    self.assertAllClose(converted_scores, [.5, .5])

  def test_build_sigmoid_score_converter(self):
    post_processing_text_proto = """
      score_converter: SIGMOID
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, score_converter = post_processing_builder.build(post_processing_config)
    self.assertEqual(score_converter.__name__, 'sigmoid_with_logit_scale')

  def test_build_softmax_score_converter(self):
    post_processing_text_proto = """
      score_converter: SOFTMAX
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, score_converter = post_processing_builder.build(post_processing_config)
    self.assertEqual(score_converter.__name__, 'softmax_with_logit_scale')

  def test_build_softmax_score_converter_with_temperature(self):
    post_processing_text_proto = """
      score_converter: SOFTMAX
      logit_scale: 2.0
    """
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, score_converter = post_processing_builder.build(post_processing_config)
    self.assertEqual(score_converter.__name__, 'softmax_with_logit_scale')

  def test_build_calibrator_with_nonempty_config(self):
    """Test that identity function used when no calibration_config specified."""
    # Calibration config maps all scores to 0.5.
    post_processing_text_proto = """
      score_converter: SOFTMAX
      calibration_config {
        function_approximation {
          x_y_pairs {
              x_y_pair {
                x: 0.0
                y: 0.5
              }
              x_y_pair {
                x: 1.0
                y: 0.5
              }}}}"""
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, calibrated_score_conversion_fn = post_processing_builder.build(
        post_processing_config)
    self.assertEqual(calibrated_score_conversion_fn.__name__,
                     'calibrate_with_function_approximation')

    def graph_fn():
      input_scores = tf.constant([1, 1], tf.float32)
      outputs = calibrated_score_conversion_fn(input_scores)
      return outputs
    calibrated_scores = self.execute_cpu(graph_fn, [])
    self.assertAllClose(calibrated_scores, [0.5, 0.5])

  def test_build_temperature_scaling_calibrator(self):
    post_processing_text_proto = """
      score_converter: SOFTMAX
      calibration_config {
        temperature_scaling_calibration {
          scaler: 2.0
          }}"""
    post_processing_config = post_processing_pb2.PostProcessing()
    text_format.Merge(post_processing_text_proto, post_processing_config)
    _, calibrated_score_conversion_fn = post_processing_builder.build(
        post_processing_config)
    self.assertEqual(calibrated_score_conversion_fn.__name__,
                     'calibrate_with_temperature_scaling_calibration')

    def graph_fn():
      input_scores = tf.constant([1, 1], tf.float32)
      outputs = calibrated_score_conversion_fn(input_scores)
      return outputs
    calibrated_scores = self.execute_cpu(graph_fn, [])
    self.assertAllClose(calibrated_scores, [0.5, 0.5])

if __name__ == '__main__':
  tf.test.main()



from http.client import UnimplementedFileMode
import numpy as np
import os
import os.path
import pathlib
import io
import time
import pickle
import matplotlib
import shutil
from matplotlib import pyplot as plt
import matplotlib.patches as patches
from math import sqrt
from tqdm import tqdm
from PIL import Image
import PIL.ImageColor as ImageColor
import PIL.ImageDraw as ImageDraw
import PIL.ImageFont as ImageFont
import pandas as pd
import random
import cv2

from sklearn.utils import shuffle
from sklearn.model_selection import train_test_split
from sklearn.model_selection import GridSearchCV
from sklearn.utils import class_weight
from sklearn.metrics import confusion_matrix, roc_curve, roc_auc_score, auc
import plotly.graph_objects as go
import plotly.express as px

import tensorflow as tf
from distutils.version import StrictVersion
if StrictVersion(tf.__version__) < StrictVersion('1.4.0'):
  raise ImportError('Please upgrade your tensorflow installation to v1.4.* or later!')

from odn.tf_ssd.object_detection.utils import label_map_util
from odn.tf_ssd.object_detection.utils import visualization_utils as vis_util
import importlib
importlib.reload(vis_util) # reflect changes in the source file immediately

class demographics():
    '''
    Provides static ploting functions for subject demographics, Such As Gender, Birth Weight, Gestational Age, etc.
    '''

    def piechart_binary(values = [50, 50], labels= ['Group 1', 'Group 2'],
        colors= ['gold', 'mediumturquoise'] ):
        fig = go.Figure(data=[go.Pie(labels=labels, values=values, textinfo='label+percent', textfont_size=20,
                                    marker=dict(colors=colors, line=dict(color='#000000', width=2)),
                                    # insidetextorientation='radial' # this will cause text rotate
                                    )])
        fig.show()

    def piechart_preterm(values):
        demographics.piechart_binary(values = values, labels= ['Preterms', 'Non-preterms'])

    def piechart_gender(values = [50, 50]):        
        demographics.piechart_binary(values = values, labels= ['male', 'female'])

    def piechart_rop(values = [50, 50]):        
        demographics.piechart_binary(values = values, labels= ['ROP', 'non-ROP'])

    def hist_gw(ga_week):
        '''
        Plot gestational age histgram in weeks
        '''
        # create the bins
        hist, bins = np.histogram(ga_week, bins = 40, range=(22,42), density = True)
        # hist = np.append(hist, 0)
        # bins = 0.5 * (bins[:-1] + bins[1:])
        fig = px.bar(x = bins[:-1], y=hist/hist.sum() * 100, labels={'x':'Gestational Age (Weeks)', 'y':'Percentage (%)'})
        # fig = px.histogram(data, x="total_bill", histnorm='probability density')
        fig.update_layout(plot_bgcolor = "white")
        fig.update_traces(marker_color='mediumturquoise', marker_line_color='rgb(8,48,107)',
                        marker_line_width=1.5, opacity=0.8)
        fig.show()

    def hist_bw(bw_kg):
        '''
        Plot birth weight histgram in kg
        '''
        # create the bins
        hist, bins = np.histogram(bw_kg, bins = 21, range=(1,2.6), density = True)
        # hist = np.insert(hist, -1, 0)
        # bins = 0.5 * (bins[:-1] + bins[1:])
        fig = px.bar(x = bins[:-1], y=hist/hist.sum() * 100, labels={'x':'Birth Weight(kg)', 'y':'Percentage (%)'})
        # fig = px.histogram(data, x="total_bill", histnorm='probability density')
        fig.update_layout(plot_bgcolor = "white")
        fig.update_traces(marker_color='mediumturquoise', marker_line_color='rgb(8,48,107)',
                        marker_line_width=1.5, opacity=0.6)
        fig.show()


class annotation():

    def get_bbox_of_circle(cx, cy, h, c):

        '''
        Parameters
        ----------
        cx, cy : center of circle
        h : image height
        c : "opticDisk" or "Macula"

        Remarks
        -------
        The fundus images are fixed in 4:3 ratio.
        On a 512-pixel-high image, macula radius is 55, optic disk radius is 35
        '''
        if (c =='OpticDisk'):
            r = 35./512*h
        elif c == 'Macula':
            r = 55./512*h
        xmin = round(cx - r)
        xmax = round(cx + r)
        ymin = round(cy -r)
        ymax = round(cy + r)
        return xmin,ymin,xmax,ymax  

    def show_anno(filepath, df, savefolder = None, drawzones = False, showimg = True):    
        '''
        Display an image with the annotations. 

        Parameters
        ----------
        filepath : image file path
        df : pandas dataframe of the annotation csv file in VIA format. 

                This is the VIA annotation format
                    filename: file name of the image
                    class: denotes the class label of the ROI (region of interest)
                    cx: image width
                    cy: image height
                    xmin: x-coordinate of the bottom left part of the ROI
                    xmax: x-coordinate of the top right part of the ROI
                    ymin: y-coordinate of the bottom left part of the ROI
                    ymax: y-coordinate of the top right part of the ROI  
                    laterality: L001 = OD, L002 = OS

        saveFolder : A target file path to save the image with annotation. 
            If None, will not save.
        showing : whether display the annotated image inline.

        '''

        # restore font size
        plt.rcParams.update({'font.size': 10})    
        
        filename = os.path.basename(filepath) # file name
        
        fig = plt.figure()
        image = plt.imread(filepath)
        plt.imshow(image)
        ax = fig.gca()
        ax.set_axis_off()

        idxs = []
        base_fn = os.path.basename(filename).lower()
        for idx, fn in enumerate(df.filename):
            base_fn2 = os.path.basename(fn).lower()
            if (base_fn == base_fn2):
                idxs.append(idx)

        # print(df.iloc[idxs])

        cx = -1
        cy = -1
        cx_m = -1
        cy_m = -1

        for _, row in df.iloc[idxs].iterrows():
            # print(row)
            xmin = row.xmin
            xmax = row.xmax
            ymin = row.ymin
            ymax = row.ymax

            width = xmax - xmin
            height = ymax - ymin

            edgecolor = 'white'
            prob = ''
            if ('prob' in df.columns):
                prob = ' ' + str(round(row['prob'], 3))
            
            # assign different color to different classes of objects
            if row['class'] == 'OpticDisk':
                cx = (xmin + xmax) / 2.0
                cy = (ymin + ymax) / 2.0
                edgecolor = 'white'
            elif row['class'] == 'Macula':
                cx_m = (xmin + xmax) / 2.0
                cy_m = (ymin + ymax) / 2.0
                edgecolor = 'lavender' #'azure' 
            
            radius = 0.5
            if (cx_m != -1):
                radius = 2*( sqrt((cx-cx_m)**2 + (cy-cy_m)**2) )            
        
            # (xmin, xmax, ymin, ymax) 
            zone1 = [cx-radius, cx+radius, cy-radius*4/3, cy+radius*4/3]
            zone2p = [cx-1.3*radius, cx+1.3*radius, cy-1.3*radius*4/3, cy+1.3*radius*4/3]
            zone2 = [cx-2*radius, cx+2*radius, cy-2*radius*4/3, cy+2*radius*4/3]

            # calculate the position of anno labels
            xy = (max(1, xmin),ymin-5)
            if (ymin < 5):
                xy = (max(1, xmin),ymax)
            ax.annotate(row['class'] + prob, xy=xy, color = edgecolor)

            # add bounding boxes to the image
            rect = patches.Rectangle((xmin,ymin), width, height, edgecolor = edgecolor, facecolor = 'none')

            ax.add_patch(rect)

        if drawzones:
            # buf = io.BytesIO()
            # fig.savefig(buf)
            # buf.seek(0)
            # image = Image.open(buf)
            # # image = Image.frombytes('RGB', fig.canvas.get_width_height(),fig.canvas.tostring_rgb())
            # annotation.draw_fundus_zones_on_image_array(image, zones = [zone1, zone2p, zone2])
            # plt.imshow(image)
            
            for zone, color in zip([zone1, zone2p, zone2], ["orange", "gold", "yellow"]):
                # (ymin, xmin, ymax, xmax)
                xmin = zone[0]
                ymin = zone[2]
                xmax = zone[1]
                ymax = zone[3]               
            
                ellipse = patches.Ellipse((cx, cy), xmax-xmin , ymax-ymin, edgecolor = color, facecolor = 'none')   # (255,255,255)
                ax.add_patch(ellipse)

        if savefolder:
            if (os.path.isdir(savefolder) == False):
                os.makedirs(savefolder)
            plt.savefig(savefolder + '/' + filename)
        
        if showimg == False:
            plt.close(fig)


    def rule_filter_rois(input_file = '../src/odn/candidate_rois.txt', 
    output_file = '../src/odn/rois.txt',
    verbose = True):

        rois_raw = pd.read_csv(input_file)
    
        if verbose:
            print('\n-------- \nInput file: ' + input_file)
            print(rois_raw.head())

        rois = rois_raw.copy()
        pd.options.mode.chained_assignment = None  # disable the "A value is trying to be set on a copy of a slice from a DataFrame" warning

        Dod = 0.1 # diameter of optic disc
        Dma = 0.16 # diameter of macula
        Dom = 0.32 # distance between optic disc and macula

        uniquefiles = set(rois['filename'].values)
        all_to_be_removed = []

        for index, fn in enumerate(uniquefiles):
            # print(fn)
            # filepath = DIR_IMAGES + '/' + fn
            rows = rois.loc[rois.filename == fn]
            ods = rows.loc[rows['class'] == 'OpticDisk']
            mas = rows.loc[rows['class'] == 'Macula']    
            to_be_removed = []
            
            if (len(ods) > 0):
                # use the max-probability optic disc
                od = ods.loc[ods['prob'].idxmax()]
                # rois.drop(ods.index[ods.index != ods['prob'].idxmax()]) # remove non-max prob bboxes
                to_be_removed = np.append(to_be_removed, ods[ods.index != ods['prob'].idxmax()].seq.values.flatten()) 
                d_od = (od['xmax'] - od['xmin']) / od['width']
                
                assert (abs(d_od - Dod) <= Dod * 0.5)
                cx_od = (od['xmax'] + od['xmin']) / 2
                cy_od = (od['ymax'] + od['ymin']) / 2
                
                if len(mas) > 0:
                
                    ma_dict={}
                    idx_least_error = -1
                    for idx,ma in mas.iterrows():            
                        cx_ma = (ma['xmax'] + ma['xmin']) / 2
                        cy_ma = (ma['ymax'] + ma['ymin']) / 2

                        dom = abs((cx_ma - cx_od) / od['width'])
                        err_dom = abs(dom - Dom)
                        if idx_least_error == -1:                
                            idx_least_error = idx
                        else:
                            if (ma_dict[idx_least_error] > err_dom):
                                idx_least_error = idx

                        ma_dict[idx] = err_dom            
                    
                    # print(ma_dict)
                    to_be_removed = np.append(to_be_removed, mas[mas.index != idx_least_error].seq.values.flatten())
                    # rois.drop(mas.index[mas.index != idx_least_error]) # only keep the region with the least error     
                    
                    ma = mas.loc[idx_least_error]
                    cx_ma = (ma['xmax'] + ma['xmin']) / 2
                    cy_ma = (ma['ymax'] + ma['ymin']) / 2           
                    dom = abs((cx_ma - cx_od) / od['width'])
                    err_dom = abs(dom - Dom)
                    
                    
                    if ('laterality' in od):
                    
                        if ((od['laterality'] == 'L001' and cx_ma >= cx_od) #  For right eye (OD) image, macula should locate on the left side of optic disc
                            or (od['laterality'] == 'L002' and cx_ma <= cx_od)): # For left eye (OS) image, macula should locate on the right side of optic disc
                            to_be_removed = np.append(to_be_removed, ma.seq)
                        
                    
                    delta_y = abs(cy_ma - cy_od) / od['width']
                    
                    if (err_dom > 0.2 or delta_y > 0.2): # max allowed error 20%
                        to_be_removed = np.append(to_be_removed, ma.seq)
                        
                
            else:
                # no optic disc is detected. Keep the maximum-prob macula 
                if (len(mas) > 0):
                    # print(mas.index[mas.index != mas['prob'].idxmax()])
                    # rois.drop(mas.index[mas.index != mas['prob'].idxmax()]) # remove non-max prob bboxes
                    to_be_removed = np.append(to_be_removed, mas[mas.index != mas['prob'].idxmax()].seq.values.flatten())
            
            rows.drop(rows.loc[rows.seq.isin(to_be_removed)].index, inplace=True)
            all_to_be_removed = np.concatenate((all_to_be_removed, to_be_removed))
            
            #d_od = row['xmax'] - row['xmin']
            #show_anno(filepath, all, './fundus_image_dataset/ground_truth/', savefile = False, showimg = True)
            #show_anno(filepath, rows, './fundus_image_dataset/odn_10e/', savefile = True, showimg = False)

        if verbose:
            print('\n-------- \nIndices of removed annos: ', all_to_be_removed)

        rois.drop(rois.loc[rois.seq.isin(all_to_be_removed)].index, inplace=True)
        rois.to_csv(output_file)

        if verbose:
            print('\n-------- \nSaved to ', output_file)
            rois= pd.read_csv(output_file)
            print(rois.head())

        return rois


    def naive_filter_rois(input_file = '../src/odn/candidate_rois.txt', 
    output_file = '../src/odn/rois_naive.txt',
    verbose = True):

        rois_raw = pd.read_csv(input_file)
    
        if verbose:
            print('\n-------- \nInput file: ' + input_file)
            print(rois_raw.head())

        rois = rois_raw.copy()
        pd.options.mode.chained_assignment = None  # disable the "A value is trying to be set on a copy of a slice from a DataFrame" warning

        Dod = 0.1 # diameter of optic disc
        Dma = 0.16 # diameter of macula
        Dom = 0.32 # distance between optic disc and macula

        uniquefiles = set(rois['filename'].values)
        all_to_be_removed = []

        for index, fn in enumerate(uniquefiles):
            # print(fn)
    
            rows = rois.loc[rois.filename == fn]
            ods = rows.loc[rows['class'] == 'OpticDisk']
            mas = rows.loc[rows['class'] == 'Macula']    
            to_be_removed = []
            
            if (len(ods) > 0):
                to_be_removed = np.append(to_be_removed, ods[ods.index != ods['prob'].idxmax()].seq.values.flatten()) 

            if (len(mas) > 0):
                to_be_removed = np.append(to_be_removed, mas[mas.index != mas['prob'].idxmax()].seq.values.flatten())
            
            rows.drop(rows.loc[rows.seq.isin(to_be_removed)].index, inplace=True)
            all_to_be_removed = np.concatenate((all_to_be_removed, to_be_removed))

        if verbose:
            print('\n-------- \nIndices of removed annos: ', all_to_be_removed)

        rois.drop(rois.loc[rois.seq.isin(all_to_be_removed)].index, inplace=True)
        rois.to_csv(output_file)

        if verbose:
            print('\n-------- \nSaved to ', output_file)
            rois= pd.read_csv(output_file)
            print(rois.head())

        return rois

    #------------------ region tf_ssd --------------------#

    def calculate_fundus_zones(classes, scores, boxes, threshold):
        '''
        Parameters
        ----------
        classes : a list of optic disc of OD, optic disc of OS, macula

        Return
        ------
        Three bbox : Zone I, Posterior Zone II, Zone II
        '''

        idxOD = -1
        idxOS = -1
        idxMacula = -1
        
        idx = 0
                        
        for c in classes:                          
            if (c == 1 and scores[idx] > threshold):
                idxOD = idx
            if (c == 2 and scores[idx] > threshold):
                idxOS = idx
            if (c == 3 and scores[idx] > threshold):
                idxMacula = idx
            idx = idx + 1
        
        cx = -1
        cy = -1
        if (idxOD >= 0):
            cx = (boxes[idxOD][3] + boxes[idxOD][1])/2.0
            cy = (boxes[idxOD][2] + boxes[idxOD][0])/2.0
                
        if (idxOS >= 0):
            cx = (boxes[idxOS][3] + boxes[idxOS][1])/2.0
            cy = (boxes[idxOS][2] + boxes[idxOS][0])/2.0
            
        if (cx == -1):
            return []    
        
        cx_m = -1
        cy_m = -1

        if (idxMacula >= 0):
            cx_m = (boxes[idxMacula][3] + boxes[idxMacula][1])/2.0
            cy_m = (boxes[idxMacula][2] + boxes[idxMacula][0])/2.0
            
        radius = 0.5
        if (cx_m != -1):
            radius = 2*( sqrt((cx-cx_m)**2 + (cy-cy_m)**2) )            
    
        # (xmin, xmax, ymin, ymax) 
        zone1 = [cx-radius, cx+radius, cy-radius*4/3, cy+radius*4/3]
        zone2p = [cx-1.3*radius, cx+1.3*radius, cy-1.3*radius*4/3, cy+1.3*radius*4/3]
        zone2 = [cx-2*radius, cx+2*radius, cy-2*radius*4/3, cy+2*radius*4/3]
        
        return [zone1, zone2p, zone2]

    def draw_fundus_zones_on_image_array(image, zones, text ='', use_normalized_coordinates=True):
        
        image_pil = Image.fromarray(np.uint8(image)).convert('RGB')
        
        draw = ImageDraw.Draw(image_pil)
        im_width, im_height = image_pil.size
        
        colors = ["orange", "gold", "yellow"]
        idx = 0
        
        for zone in zones:
            # (ymin, xmin, ymax, xmax)
            xmin = zone[0]
            ymin = zone[2]
            xmax = zone[1]
            ymax = zone[3]
            if use_normalized_coordinates:
                (left, right, top, bottom) = (xmin * im_width, xmax * im_width, ymin * im_height, ymax * im_height)
            else:
                (left, right, top, bottom) = (xmin, xmax, ymin, ymax)      
        
            draw.ellipse((int(left), int(top), int(right),int(bottom)), fill = None, outline =colors[idx])   # (255,255,255)
            
            idx+=1
            
        try:
            font = ImageFont.truetype('arial.ttf', 12)
        except IOError:
            font = ImageFont.load_default()

        draw.text(
            (20,20),
            text,
            fill=(255,255,255),
            font=font)    
        
        np.copyto(image, np.array(image_pil))

    def load_tf_graph(ckpt_path = '../src/odn/tf_ssd/export/frozen_inference_graph.pb',
                 label_path = '../src/odn/tf_ssd/fundus_label_map.pbtxt', 
                 num_classes = 2, verbose = True):
        '''
        Load tf graph from checkpoint
        
        Parameters
        ----------
        ckpt_path : path to the model checkpoint file. e.g., '../src/odn/tf_ssd/export/frozen_inference_graph.pb', 'ssd_mobilenet_v1_coco_2017_11_17/frozen_inference_graph.pb'
        label_path : path to label file. Should follow this json format:
            item {
            id: 1
            name: 'OpticDisk'
            }

            item {
            id: 2
            name: 'Macula'
            }
        
        Returns
        -------    
        detection_graph : a TensorFlow computation, represented as a dataflow graph.
        '''

        detection_graph = tf.Graph()
        with detection_graph.as_default():
            od_graph_def = tf.GraphDef()
            with tf.gfile.GFile(ckpt_path, 'rb') as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name='')


        label_map = label_map_util.load_labelmap(label_path)
        categories = label_map_util.convert_label_map_to_categories(label_map, max_num_classes=num_classes, use_display_name=True)
        category_index = label_map_util.create_category_index(categories)
        
        if verbose:
            print('category_index: ', category_index)
        
        return detection_graph, category_index


    def tf_batch_object_detection(detection_graph, category_index, FILES, 
                            target_folder, log_file, 
                            display = False, savefile = True, 
                            IMAGE_SIZE = (24, 18), threshold = 0.2, 
                            new_img_width = None,
                            fontsize = 24):    
        if savefile:
            os.makedirs(target_folder, exist_ok=True) # create the target folder if not exist
        
        if (log_file is not None and log_file != ''):
            if os.path.exists(log_file):
                os.remove(log_file)
            try:
                open(log_file, 'w')
            except OSError:
                print('Warning: log file path ', log_file, ' is invalid')
                log_file = None
        
        with detection_graph.as_default():
            with tf.Session(graph=detection_graph) as sess:
                # Definite input and output Tensors for detection_graph
                image_tensor = detection_graph.get_tensor_by_name('image_tensor:0')
                # Each box represents a part of the image where a particular object was detected.
                detection_boxes = detection_graph.get_tensor_by_name('detection_boxes:0')
                # Each score represent how level of confidence for each of the objects.
                # Score is shown on the result image, together with the class label.
                detection_scores = detection_graph.get_tensor_by_name('detection_scores:0')
                detection_classes = detection_graph.get_tensor_by_name('detection_classes:0')
                num_detections = detection_graph.get_tensor_by_name('num_detections:0')
                with tqdm(total=len(FILES)) as pbar:
                    for image_path in FILES:
                        image = Image.open(image_path)
                        if new_img_width > 0:
                            image.thumbnail((new_img_width,new_img_width)) # Image.ANTIALIAS
                        # the array based representation of the image will be used later in order to prepare the result image with boxes and labels on it.
                        image_np = load_image_into_numpy_array(image)
                        # Expand dimensions since the model expects images to have shape: [1, None, None, 3]
                        image_np_expanded = np.expand_dims(image_np, axis=0)
                        # Actual detection.
                        (boxes, scores, classes, num) = sess.run(
                            [detection_boxes, detection_scores, detection_classes, num_detections],
                            feed_dict={image_tensor: image_np_expanded})
                            
                        # get image dims
                        (im_width, im_height) = image.size
                            
                        # special treatment: only keep one opticdisk: OD or OS
                        idxOD = -1
                        idxOS = -1
                        idxMacula = -1
                        probMacula = 0
                        idx = 0
                        
                        # log object detection info
                        info = os.path.basename(image_path) + ' '                    
                        
                        for c in classes[0]:                          
                            
                            info = info + str(int(c)) + ' ' + str(round(scores[0][idx],3)) + ' '
                            
                            if (c == 1):
                                idxOD = idx
                            if (c == 2):
                                idxOS = idx
                            if (c == 3):
                                if idxMacula == -1 or probMacula < scores[0][idx]:
                                    if idxMacula > 0:
                                        scores[0][idxMacula] = 0.0 
                                    idxMacula = idx
                                    probMacula = scores[0][idx]
                                    
                            idx = idx + 1
                        
                        if display: 
                            print('Top Objects:\n', 'boxes = ', boxes[0], '\nscores = ', scores[0], '\nclasses = ', classes[0])
                        
                    
                        ###### RULE1 #######
                        # Discard oversized macula candidates; Judge the distance between macula and OpticDisk                       
                        # On a 512-pixel-high image, macula radius is 55, optic disk radius is 35
                        # The boxes object is a 2 dimensional numpy array of [N, 4]: (ymin, xmin, ymax, xmax). The coordinates are in normalized format between [0, 1].
                        m_w = abs(boxes[0][idxMacula][3] - boxes[0][idxMacula][1])
                        m_h = abs(boxes[0][idxMacula][2] - boxes[0][idxMacula][0])
                        m_cx = abs(boxes[0][idxMacula][3] + boxes[0][idxMacula][1])/2.0
                        m_cy = abs(boxes[0][idxMacula][2] + boxes[0][idxMacula][0])/2.0
                        
                        # print(m_w, m_h, m_cx, m_cy)
                        
                        if m_w > 0.3 or m_h > 0.3:
                            scores[0][idxMacula] = 0.0
                        
                        
                        ##### RULE2: Optic Disk OD/OS Judgment ######
                        od_cx = abs(boxes[0][idxOD][3] + boxes[0][idxOD][1])/2.0
                        od_cy = abs(boxes[0][idxOD][2] + boxes[0][idxOD][0])/2.0
                                
                        os_cx = abs(boxes[0][idxOS][3] + boxes[0][idxOS][1])/2.0
                        os_cy = abs(boxes[0][idxOS][2] + boxes[0][idxOS][0])/2.0
                                
                        # OpticDisk near the left rim is OS
                        if (scores[0][idxOD] > threshold and scores[0][idxOD]>scores[0][idxOS] and od_cx < 0.2): # 2.0*35*2/512
                            scores[0][idxOS] = max(scores[0][idxOS], scores[0][idxOD])
                            scores[0][idxOD] = 0
                            boxes[0][idxOS] = boxes[0][idxOD] # ?üD?bbox
                            #classes[0][idxOD] = 2 # set as OS
                                
                        # OpticDisk near the right rim is OD
                        if (scores[0][idxOS] > threshold  and scores[0][idxOD]<scores[0][idxOS] and os_cx > 0.8):
                            scores[0][idxOD] = max(scores[0][idxOS], scores[0][idxOD])
                            scores[0][idxOS] = 0
                            boxes[0][idxOD] = boxes[0][idxOS] # ?üD?bbox
                            #classes[0][idxOS] = 1 # set as OD
                            
                        
                        ##### RULE3: Judge relative positions of macula and OpticDisk
                        
                        # reload boxes info
                        od_cx = abs(boxes[0][idxOD][3] + boxes[0][idxOD][1])/2.0
                        od_cy = abs(boxes[0][idxOD][2] + boxes[0][idxOD][0])/2.0
                                
                        os_cx = abs(boxes[0][idxOS][3] + boxes[0][idxOS][1])/2.0
                        os_cy = abs(boxes[0][idxOS][2] + boxes[0][idxOS][0])/2.0
                        
                        if (scores[0][idxMacula] > threshold):
                            if (scores[0][idxOD] > threshold and scores[0][idxOD] > scores[0][idxOS] and m_cx > od_cx ):
                                scores[0][idxOS] = max(scores[0][idxOS], scores[0][idxOD])
                                scores[0][idxOD] = 0
                                boxes[0][idxOS] = boxes[0][idxOD]
                            if (scores[0][idxOS] > threshold and scores[0][idxOD] < scores[0][idxOS] and m_cx < os_cx):
                                scores[0][idxOD] = max(scores[0][idxOS], scores[0][idxOD])
                                scores[0][idxOS] = 0
                                boxes[0][idxOD] = boxes[0][idxOS] # update bbox
                        
                        
                        ##### RULE4: Keep the bigger probality of OpticDisk
                        
                        if (scores[0][idxOD] > scores[0][idxOS]):                        
                            scores[0][idxOS] = 0.0                        
                        if (scores[0][idxOD] < scores[0][idxOS]):
                            scores[0][idxOD] = 0.0                                    
                            

                        info += ' ; ' 
                        idx = 0
                        label = ''
                        for c in classes[0]:                        
                            info = info + str(round(scores[0][idx],3)) + ' '                        
                            if (scores[0][idx] > threshold):
                                label += category_index[c]['name'] + ' ' + str(round(scores[0][idx],3)) + '  '
                            idx += 1
                                        
                        zones = annotation.calculate_fundus_zones(classes[0], scores[0], boxes[0], threshold)
                        annotation.draw_fundus_zones_on_image_array(image_np, zones = zones, text=label)
                        
                        # Visualization of the results of a detection.
                        vis_util.visualize_boxes_and_labels_on_image_array(
                            image_np,
                            np.squeeze(boxes),
                            np.squeeze(classes).astype(np.int32),
                            np.squeeze(scores),
                            category_index,
                            use_normalized_coordinates=True,
                            line_thickness=1, # 3   
                            max_boxes_to_draw=2, # 3
                            min_score_thresh=threshold,
                            fontsize = fontsize)
                        fig = plt.figure(figsize=IMAGE_SIZE)
                        if (display):
                            plt.imshow(image_np)
                        if(savefile):
                            plt.imsave(os.path.join(target_folder, os.path.basename(image_path)), image_np)
                            # plt.annotate(info, (0, 0), color='b', weight='bold', fontsize=12, ha='left', va='top')
                            # print(info)
                            plt.close(fig)
                            
                        if (log_file is not None and log_file != ''): # the validity of log_file is already checked in the beginning
                            with open(log_file, "a") as myfile:
                                myfile.write(info + '\n')                        
                        pbar.update(1)                        


class dataset():

    def synthesize_anno(label_file, 
    dir_images, 
    dir_output = None, 
    drawzones = False,   
    verbose = True,
    display = 5 ):
        '''
        Synthesize the ground-truth images with annotations.

        Parameters
        ----------
        label_file : a csv file containing all image path and ROIs. e.g.,  '../data/fundus/all_labels.csv'

            The label file extends the VIA annotation format:

                filename: file name of the image
                class: denotes the class label of the ROI (region of interest)
                cx: image width
                cy: image height
                xmin: x-coordinate of the bottom left part of the ROI
                xmax: x-coordinate of the top right part of the ROI
                ymin: y-coordinate of the bottom left part of the ROI
                ymax: y-coordinate of the top right part of the ROI  
                laterality: L001 = OD, L002 = OS

        dir_images : The folder path containing all images, e.g.,  '../data/fundus/images/'

        dir_output : The generated images will be saved to this folder. e.g., '../data/fundus/ground_truth/'.
            If None, will not output images.
        '''
        
        df = pd.read_csv(label_file)

        if verbose:
            print('\n----------- \nContent of ', label_file)
            print(df.head())

        unique_files = list(set(df['filename'].values))
        if verbose:
            print('\n----------- \nUnique image files: ', len(unique_files))

        if verbose:
            print('\n----------- \ndistribution of ROI labels:')
            print( df['class'].value_counts() )
        
        i = 0
        for f in unique_files:            
            filepath = dir_images + '/' + f
            annotation.show_anno (filepath, df, dir_output, 
            drawzones = drawzones, showimg = (i < display) )
            i = i+1


    def split(label_file, dir_images, train_output, test_output,
    test_size = 0.2, verbose = True):
        '''
        Split the dataset into training and test sets.

        Parameters
        ----------
        label_file : a csv file containing all image path and ROIs. e.g.,  '../data/fundus/all_labels.csv'

            The label file extends the VIA annotation format:

                filename: file name of the image
                class: denotes the class label of the ROI (region of interest)
                cx: image width
                cy: image height
                xmin: x-coordinate of the bottom left part of the ROI
                xmax: x-coordinate of the top right part of the ROI
                ymin: y-coordinate of the bottom left part of the ROI
                ymax: y-coordinate of the top right part of the ROI  
                laterality: L001 = OD, L002 = OS

        dir_images : The folder path containing all images, e.g.,  '../data/fundus/images/'

        train_output, test_output : Target paths of generated train and test csv files.

        test_size : percentage of test set, in the range (0, 1.0)
        '''

        df = pd.read_csv(label_file)

        if verbose:
            print('\n----------- \nContent of ', label_file)
            print(df.head())

        unique_files = list(set(df['filename'].values))
        if verbose:
            print('\n----------- \nUnique image files: ', len(unique_files))

        m = len(unique_files)
        m_t = round(m * test_size)
        test_files = random.choices(unique_files, k = m_t)

        # Split to training set and test set

        test_set = df.loc[df['filename'].isin(test_files)].copy().sort_values(by=['filename']).reset_index(drop=True)
        train_set = df.loc[df['filename'].isin(test_files) == False].copy().sort_values(by=['filename']).reset_index(drop=True)

        # training set

        data = pd.DataFrame()
        data['format'] = train_set['filename']

        for i in range(data.shape[0]):
            # the path is relative to /src/odn/..py
            data['format'][i] = '../' + dir_images + data['format'][i] + ',' + str(train_set['xmin'][i]) + ',' + str(train_set['ymin'][i]) + ',' + str(train_set['xmax'][i]) + ',' + str(train_set['ymax'][i]) + ',' + train_set['class'][i]
            # print(data['format'][i])
            
        data.to_csv(train_output, header=None, index=None, sep=' ')


        # test set. We only need the filenames. The bbox of ROIs are not used.

        data = pd.DataFrame()
        data['format'] = test_set['filename']

        for i in range(data.shape[0]):
            data['format'][i] = '../' + dir_images + data['format'][i] + ',' + str(test_set['xmin'][i]) + ',' + str(test_set['ymin'][i]) + ',' + str(test_set['xmax'][i]) + ',' + str(test_set['ymax'][i]) + ',' + test_set['class'][i]

        data.to_csv(test_output, header=None, index=None, sep=' ')

    def add_new_row(img_, bboxes_, row, idx, 
    old_image_path_seg = '/images', 
    new_image_path_seg = '/images_public'):
    
        new_path = row[0].replace(old_image_path_seg, new_image_path_seg, 1) # .replace('.jpg', '.' + str(idx) + '.jpg', 1)
        status = cv2.imwrite(new_path, img_[:,:,::-1]) # bgr -> rgb
        
        s = new_path + \
        ',' + str(round( bboxes_[0][0], 1)) + \
        ',' + str(round( bboxes_[0][1], 1)) + \
        ',' + str(round( bboxes_[0][2], 1)) + \
        ',' + str(round( bboxes_[0][3], 1)) + \
        ',' + row[5]
        
        return s

    def deidentify(input_file = '../data/fundus/train.txt',
            output_file = '../data/fundus/train_public.txt', 
            old_image_path_seg = '/images', 
            new_image_path_seg = '/images_public'):

        # thanks to "https://github.com/Paperspace/DataAugmentationForObjectDetection"

        # %run ../src/odn/data_aug/data_aug.py
        # %run ../src/odn/data_aug/bbox_util.py

        # from data_aug.data_aug import *
        # from data_aug.bbox_util import *

        df = pd.read_csv(input_file, header=None, sep=',')

        datax = []

        DEID = True
        AUG_FLIP = False # always disabled for test
        AUG_COMBO = False # always disabled for test

        for index, row in df.iterrows():
            
            path = row[0][1:]
            # print([[row[1],row[2],row[3],row[4]]])
            img = cv2.imread(path)[:,:,::-1] #opencv loads images in bgr. the [:,:,::-1] does bgr -> rgb    
            
            bboxes = np.array([[row[1],row[2],row[3],row[4]]], dtype = float)
            
            # 
            # De-identify
            
            if DEID:
            
                w = img.shape[1]
                h = img.shape[0]

                # crop the right bottom section to detect text
                crop_img = img[round(h*0.93) : h, round(w * 0.93): w]
                gray = cv2.cvtColor(crop_img, cv2.COLOR_BGR2GRAY) 
                _, thresh = cv2.threshold(gray, 127, 255,  cv2.THRESH_BINARY_INV)  # cv2.THRESH_OTSU |

                if (thresh.min() < 127):

                    img[ round(h*0.8) : h, round(w * 0.8): w] = 0 # de-identify

                    if (index < 5):
                        plt.figure()
                        plt.imshow(img)
                        plt.title( path )
                        plt.show()

                        plt.figure()
                        plt.imshow(gray)
                        plt.title( str(thresh.mean()) )
                        plt.show()

                s = dataset.add_new_row(
                    img, bboxes, row, 0, 
                    old_image_path_seg,
                    new_image_path_seg)
                datax.append(s)

            # 
            # DATA AUG
            
            if AUG_FLIP:
                
                raise NotImplementedError(__class__.__name__ + ' TODO: 把多个bbox和标签放到一起做aug操作')

                # FLIP
                img_, bboxes_ = RandomHorizontalFlip(1)(img.copy(), bboxes.copy())
                s = add_new_row(img_, bboxes_, row, 1)
                datax.append(s)
            
            if AUG_COMBO: # FLIP + SCALE + TRANSLATE + ROTATE   
                
                raise NotImplementedError(__class__.__name__ + ' TODO: 把多个bbox和标签放到一起做aug操作')

                for idx in range(2, 7):
                    try:
                        img_, bboxes_ = RandomHorizontalFlip(0.5)(img.copy(), bboxes.copy())
                        img_, bboxes_ = RandomScale(0.1, diff = False)(img_, bboxes_)
                        img_, bboxes_ = RandomTranslate(0.1, diff = False)(img_, bboxes_)
                        img_, bboxes_ = RandomRotate(8)(img_, bboxes_)

                        s = add_new_row(img_, bboxes_, row, idx)
                        datax.append(s)

                    except Exception as err:
                        # print(row)
                        pass
            
                # show the first N examples
                if (index < 5):
                    plotted_img = draw_rect(img_, bboxes_)
                    plt.figure()
                    plt.imshow(plotted_img)
                    plt.title(row[5])
                    plt.show()
                
        sx = ''
        for d in datax:
            sx = sx + d + '\n'

        with open(output_file, "w") as f:
            f.write(sx)
            
        # finally, combine trainx.txt with train.txt

    # NOTE: images in the target folder will be overwritten if prefix is empty
    def batch_resize_fundus_image(folder, target, prefix = '', w=480, h=360):
        os.makedirs(target, exist_ok=True)
        
        imgExts = ["png", "bmp", "jpg"]
        for root, dirs, files in os.walk(folder):
            for f in files:
                ext = f[-3:].lower()
                if ext not in imgExts:
                    continue
                if prefix!='' and prefix is not None and f.startswith(prefix):
                    continue
                resize_fundus_image(root, f, target = target, prefix=prefix, w = w, h = h)

    def split_images_by_laterality(source_dir, target_dir, json_annos):
        '''
        Parameters
        ----------
        json_annos : a json object that contains a list of dict objects. 
            We will use its 'laterality' key to sperate images into OD and OS.
        '''
            
        IMAGEDIR = source_dir #'../data/fundus/images/'
        L2DIR = target_dir # '../data/fundus/L2'

        if os.path.exists(L2DIR):
            shutil.rmtree(L2DIR)
        os.makedirs(L2DIR)
        os.makedirs(L2DIR+'/L001')
        os.makedirs(L2DIR+'/L002')

        for idx, item in enumerate(json_annos):
            if 'laterality' in item.keys():
                lcode = item['laterality']    
                if lcode and not lcode.isspace():
                    file = item['filename']
                    shutil.copyfile(os.path.join(IMAGEDIR,file), os.path.join(L2DIR,lcode,file))

    def expand_images_by_flipping(imagedir):
        '''
        Create a flip copy for each image. The new image is named as filename_FLIP.jpg
        '''

        for root, dirs, files in os.walk(imagedir):
            for f in files:
                if f.endswith('.jpg'):
                    im = Image.open(os.path.join(root, f)).transpose(Image.FLIP_LEFT_RIGHT)                
                    im.save(os.path.join(root, f.replace('.jpg','_FLIP.jpg')))

def load_image_into_numpy_array(image):
    (im_width, im_height) = image.size
    return np.array(image.getdata()).reshape((im_height, im_width, 3)).astype(np.uint8)
    
# All fundus images are 4:3 ratio. 
def resize_fundus_image(root, f, target, prefix='', w=480, h=360):
    filePath = os.path.join(root, f)
    newfilePath = os.path.join(target, prefix + f)    
    im = Image.open(filePath).resize((w,h)) 
    im.save(newfilePath)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: object_detection/protos/matcher.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ..protos import argmax_matcher_pb2 as object__detection_dot_protos_dot_argmax__matcher__pb2
from ..protos import bipartite_matcher_pb2 as object__detection_dot_protos_dot_bipartite__matcher__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='object_detection/protos/matcher.proto',
  package='object_detection.protos',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n%object_detection/protos/matcher.proto\x12\x17object_detection.protos\x1a,object_detection/protos/argmax_matcher.proto\x1a/object_detection/protos/bipartite_matcher.proto\"\xa4\x01\n\x07Matcher\x12@\n\x0e\x61rgmax_matcher\x18\x01 \x01(\x0b\x32&.object_detection.protos.ArgMaxMatcherH\x00\x12\x46\n\x11\x62ipartite_matcher\x18\x02 \x01(\x0b\x32).object_detection.protos.BipartiteMatcherH\x00\x42\x0f\n\rmatcher_oneof'
  ,
  dependencies=[object__detection_dot_protos_dot_argmax__matcher__pb2.DESCRIPTOR,object__detection_dot_protos_dot_bipartite__matcher__pb2.DESCRIPTOR,])




_MATCHER = _descriptor.Descriptor(
  name='Matcher',
  full_name='object_detection.protos.Matcher',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='argmax_matcher', full_name='object_detection.protos.Matcher.argmax_matcher', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bipartite_matcher', full_name='object_detection.protos.Matcher.bipartite_matcher', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='matcher_oneof', full_name='object_detection.protos.Matcher.matcher_oneof',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=162,
  serialized_end=326,
)

_MATCHER.fields_by_name['argmax_matcher'].message_type = object__detection_dot_protos_dot_argmax__matcher__pb2._ARGMAXMATCHER
_MATCHER.fields_by_name['bipartite_matcher'].message_type = object__detection_dot_protos_dot_bipartite__matcher__pb2._BIPARTITEMATCHER
_MATCHER.oneofs_by_name['matcher_oneof'].fields.append(
  _MATCHER.fields_by_name['argmax_matcher'])
_MATCHER.fields_by_name['argmax_matcher'].containing_oneof = _MATCHER.oneofs_by_name['matcher_oneof']
_MATCHER.oneofs_by_name['matcher_oneof'].fields.append(
  _MATCHER.fields_by_name['bipartite_matcher'])
_MATCHER.fields_by_name['bipartite_matcher'].containing_oneof = _MATCHER.oneofs_by_name['matcher_oneof']
DESCRIPTOR.message_types_by_name['Matcher'] = _MATCHER
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Matcher = _reflection.GeneratedProtocolMessageType('Matcher', (_message.Message,), {
  'DESCRIPTOR' : _MATCHER,
  '__module__' : 'object_detection.protos.matcher_pb2'
  # @@protoc_insertion_point(class_scope:object_detection.protos.Matcher)
  })
_sym_db.RegisterMessage(Matcher)


# @@protoc_insertion_point(module_scope)



import os
import cv2
import numpy as np
import sys
import pickle
from optparse import OptionParser
import time
from keras_frcnn import config
import keras_frcnn.resnet as nn
from keras import backend as K
from keras.layers import Input
from keras.models import Model
from keras_frcnn import roi_helpers
from keras_frcnn import data_generators
from sklearn.metrics import average_precision_score


def get_map(pred, gt, f):
	T = {}
	P = {}
	fx, fy = f

	for bbox in gt:
		bbox['bbox_matched'] = False

	pred_probs = np.array([s['prob'] for s in pred])
	box_idx_sorted_by_prob = np.argsort(pred_probs)[::-1]

	for box_idx in box_idx_sorted_by_prob:
		pred_box = pred[box_idx]
		pred_class = pred_box['class']
		pred_x1 = pred_box['x1']
		pred_x2 = pred_box['x2']
		pred_y1 = pred_box['y1']
		pred_y2 = pred_box['y2']
		pred_prob = pred_box['prob']
		if pred_class not in P:
			P[pred_class] = []
			T[pred_class] = []
		P[pred_class].append(pred_prob)
		found_match = False

		for gt_box in gt:
			gt_class = gt_box['class']
			gt_x1 = gt_box['x1']/fx
			gt_x2 = gt_box['x2']/fx
			gt_y1 = gt_box['y1']/fy
			gt_y2 = gt_box['y2']/fy
			gt_seen = gt_box['bbox_matched']
			if gt_class != pred_class:
				continue
			if gt_seen:
				continue
			iou = data_generators.iou((pred_x1, pred_y1, pred_x2, pred_y2), (gt_x1, gt_y1, gt_x2, gt_y2))
			if iou >= 0.5:
				found_match = True
				gt_box['bbox_matched'] = True
				break
			else:
				continue

		T[pred_class].append(int(found_match))

	for gt_box in gt:
		if not gt_box['bbox_matched'] and not gt_box['difficult']:
			if gt_box['class'] not in P:
				P[gt_box['class']] = []
				T[gt_box['class']] = []

			T[gt_box['class']].append(1)
			P[gt_box['class']].append(0)

	#import pdb
	#pdb.set_trace()
	return T, P

sys.setrecursionlimit(40000)

parser = OptionParser()

parser.add_option("-p", "--path", dest="test_path", help="Path to test data.")
parser.add_option("-n", "--num_rois", dest="num_rois",
				help="Number of ROIs per iteration. Higher means more memory use.", default=32)
parser.add_option("--config_filename", dest="config_filename", help=
				"Location to read the metadata related to the training (generated when training).",
				default="config.pickle")
parser.add_option("-o", "--parser", dest="parser", help="Parser to use. One of simple or pascal_voc",
				default="pascal_voc"),

(options, args) = parser.parse_args()

if not options.test_path:   # if filename is not given
	parser.error('Error: path to test data must be specified. Pass --path to command line')


if options.parser == 'pascal_voc':
	from keras_frcnn.pascal_voc_parser import get_data
elif options.parser == 'simple':
	from keras_frcnn.simple_parser import get_data
else:
	raise ValueError("Command line option parser must be one of 'pascal_voc' or 'simple'")

config_output_filename = options.config_filename

with open(config_output_filename, 'r') as f_in:
	C = pickle.load(f_in)

# turn off any data augmentation at test time
C.use_horizontal_flips = False
C.use_vertical_flips = False
C.rot_90 = False

img_path = options.test_path


def format_img(img, C):
	img_min_side = float(C.im_size)
	(height,width,_) = img.shape
	
	if width <= height:
		f = img_min_side/width
		new_height = int(f * height)
		new_width = int(img_min_side)
	else:
		f = img_min_side/height
		new_width = int(f * width)
		new_height = int(img_min_side)
	fx = width/float(new_width)
	fy = height/float(new_height)
	img = cv2.resize(img, (new_width, new_height), interpolation=cv2.INTER_CUBIC)
	img = img[:, :, (2, 1, 0)]
	img = img.astype(np.float32)
	img[:, :, 0] -= C.img_channel_mean[0]
	img[:, :, 1] -= C.img_channel_mean[1]
	img[:, :, 2] -= C.img_channel_mean[2]
	img /= C.img_scaling_factor
	img = np.transpose(img, (2, 0, 1))
	img = np.expand_dims(img, axis=0)
	return img, fx, fy


class_mapping = C.class_mapping

if 'bg' not in class_mapping:
	class_mapping['bg'] = len(class_mapping)

class_mapping = {v: k for k, v in class_mapping.iteritems()}
print(class_mapping)
class_to_color = {class_mapping[v]: np.random.randint(0, 255, 3) for v in class_mapping}
C.num_rois = int(options.num_rois)

if K.image_dim_ordering() == 'th':
	input_shape_img = (3, None, None)
	input_shape_features = (1024, None, None)
else:
	input_shape_img = (None, None, 3)
	input_shape_features = (None, None, 1024)


img_input = Input(shape=input_shape_img)
roi_input = Input(shape=(C.num_rois, 4))
feature_map_input = Input(shape=input_shape_features)

# define the base network (resnet here, can be VGG, Inception, etc)
shared_layers = nn.nn_base(img_input, trainable=True)

# define the RPN, built on the base layers
num_anchors = len(C.anchor_box_scales) * len(C.anchor_box_ratios)
rpn_layers = nn.rpn(shared_layers, num_anchors)

classifier = nn.classifier(feature_map_input, roi_input, C.num_rois, nb_classes=len(class_mapping), trainable=True)

model_rpn = Model(img_input, rpn_layers)
model_classifier_only = Model([feature_map_input, roi_input], classifier)

model_classifier = Model([feature_map_input, roi_input], classifier)

model_rpn.load_weights(C.model_path, by_name=True)
model_classifier.load_weights(C.model_path, by_name=True)

model_rpn.compile(optimizer='sgd', loss='mse')
model_classifier.compile(optimizer='sgd', loss='mse')

all_imgs, _, _ = get_data(options.test_path)
test_imgs = [s for s in all_imgs if s['imageset'] == 'test']


T = {}
P = {}
for idx, img_data in enumerate(test_imgs):
	print('{}/{}'.format(idx,len(test_imgs)))
	st = time.time()
	filepath = img_data['filepath']

	img = cv2.imread(filepath)

	X, fx, fy = format_img(img, C)

	if K.image_dim_ordering() == 'tf':
		X = np.transpose(X, (0, 2, 3, 1))

	# get the feature maps and output from the RPN
	[Y1, Y2, F] = model_rpn.predict(X)

	R = roi_helpers.rpn_to_roi(Y1, Y2, C, K.image_dim_ordering(), overlap_thresh=0.7)

	# convert from (x1,y1,x2,y2) to (x,y,w,h)
	R[:, 2] -= R[:, 0]
	R[:, 3] -= R[:, 1]

	# apply the spatial pyramid pooling to the proposed regions
	bboxes = {}
	probs = {}

	for jk in range(R.shape[0] // C.num_rois + 1):
		ROIs = np.expand_dims(R[C.num_rois * jk:C.num_rois * (jk + 1), :], axis=0)
		if ROIs.shape[1] == 0:
			break

		if jk == R.shape[0] // C.num_rois:
			# pad R
			curr_shape = ROIs.shape
			target_shape = (curr_shape[0], C.num_rois, curr_shape[2])
			ROIs_padded = np.zeros(target_shape).astype(ROIs.dtype)
			ROIs_padded[:, :curr_shape[1], :] = ROIs
			ROIs_padded[0, curr_shape[1]:, :] = ROIs[0, 0, :]
			ROIs = ROIs_padded

		[P_cls, P_regr] = model_classifier_only.predict([F, ROIs])

		for ii in range(P_cls.shape[1]):

			if np.argmax(P_cls[0, ii, :]) == (P_cls.shape[2] - 1):
				continue

			cls_name = class_mapping[np.argmax(P_cls[0, ii, :])]

			if cls_name not in bboxes:
				bboxes[cls_name] = []
				probs[cls_name] = []

			(x, y, w, h) = ROIs[0, ii, :]

			cls_num = np.argmax(P_cls[0, ii, :])
			try:
				(tx, ty, tw, th) = P_regr[0, ii, 4 * cls_num:4 * (cls_num + 1)]
				tx /= C.classifier_regr_std[0]
				ty /= C.classifier_regr_std[1]
				tw /= C.classifier_regr_std[2]
				th /= C.classifier_regr_std[3]
				x, y, w, h = roi_helpers.apply_regr(x, y, w, h, tx, ty, tw, th)
			except:
				pass
			bboxes[cls_name].append([16 * x, 16 * y, 16 * (x + w), 16 * (y + h)])
			probs[cls_name].append(np.max(P_cls[0, ii, :]))

	all_dets = []

	for key in bboxes:
		bbox = np.array(bboxes[key])

		new_boxes, new_probs = roi_helpers.non_max_suppression_fast(bbox, np.array(probs[key]), overlap_thresh=0.5)
		for jk in range(new_boxes.shape[0]):
			(x1, y1, x2, y2) = new_boxes[jk, :]
			det = {'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2, 'class': key, 'prob': new_probs[jk]}
			all_dets.append(det)


	print('Elapsed time = {}'.format(time.time() - st))
	t, p = get_map(all_dets, img_data['bboxes'], (fx, fy))
	for key in t.keys():
		if key not in T:
			T[key] = []
			P[key] = []
		T[key].extend(t[key])
		P[key].extend(p[key])
	all_aps = []
	for key in T.keys():
		ap = average_precision_score(T[key], P[key])
		print('{} AP: {}'.format(key, ap))
		all_aps.append(ap)
	print('mAP = {}'.format(np.mean(np.array(all_aps))))
	#print(T)
	#print(P)


# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for ssd_mobilenet_v3_feature_extractor."""
import unittest
import tensorflow.compat.v1 as tf

from ..models import ssd_mobilenet_v3_feature_extractor
from ..models import ssd_mobilenet_v3_feature_extractor_testbase
from ..utils import tf_version


@unittest.skipIf(tf_version.is_tf2(), 'Skipping TF1.X only test.')
class SsdMobilenetV3LargeFeatureExtractorTest(
    ssd_mobilenet_v3_feature_extractor_testbase
    ._SsdMobilenetV3FeatureExtractorTestBase):

  def _get_input_sizes(self):
    """Return first two input feature map sizes."""
    return [672, 480]

  def _create_feature_extractor(self,
                                depth_multiplier,
                                pad_to_multiple,
                                use_explicit_padding=False,
                                use_keras=False):
    """Constructs a new Mobilenet V3-Large feature extractor.

    Args:
      depth_multiplier: float depth multiplier for feature extractor
      pad_to_multiple: the nearest multiple to zero pad the input height and
        width dimensions to.
      use_explicit_padding: use 'VALID' padding for convolutions, but prepad
        inputs so that the output dimensions are the same as if 'SAME' padding
        were used.
      use_keras: if True builds a keras-based feature extractor, if False builds
        a slim-based one.

    Returns:
      an ssd_meta_arch.SSDFeatureExtractor object.
    """
    min_depth = 32
    return (
        ssd_mobilenet_v3_feature_extractor.SSDMobileNetV3LargeFeatureExtractor(
            False,
            depth_multiplier,
            min_depth,
            pad_to_multiple,
            self.conv_hyperparams_fn,
            use_explicit_padding=use_explicit_padding))


@unittest.skipIf(tf_version.is_tf2(), 'Skipping TF1.X only test.')
class SsdMobilenetV3SmallFeatureExtractorTest(
    ssd_mobilenet_v3_feature_extractor_testbase
    ._SsdMobilenetV3FeatureExtractorTestBase):

  def _get_input_sizes(self):
    """Return first two input feature map sizes."""
    return [288, 288]

  def _create_feature_extractor(self,
                                depth_multiplier,
                                pad_to_multiple,
                                use_explicit_padding=False,
                                use_keras=False):
    """Constructs a new Mobilenet V3-Small feature extractor.

    Args:
      depth_multiplier: float depth multiplier for feature extractor
      pad_to_multiple: the nearest multiple to zero pad the input height and
        width dimensions to.
      use_explicit_padding: use 'VALID' padding for convolutions, but prepad
        inputs so that the output dimensions are the same as if 'SAME' padding
        were used.
      use_keras: if True builds a keras-based feature extractor, if False builds
        a slim-based one.

    Returns:
      an ssd_meta_arch.SSDFeatureExtractor object.
    """
    min_depth = 32
    return (
        ssd_mobilenet_v3_feature_extractor.SSDMobileNetV3SmallFeatureExtractor(
            False,
            depth_multiplier,
            min_depth,
            pad_to_multiple,
            self.conv_hyperparams_fn,
            use_explicit_padding=use_explicit_padding))


if __name__ == '__main__':
  tf.test.main()



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Mobilenet Base Class."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import collections
import contextlib
import copy
import os

import tensorflow.compat.v1 as tf
import tf_slim as slim


@slim.add_arg_scope
def apply_activation(x, name=None, activation_fn=None):
  return activation_fn(x, name=name) if activation_fn else x


def _fixed_padding(inputs, kernel_size, rate=1):
  """Pads the input along the spatial dimensions independently of input size.

  Pads the input such that if it was used in a convolution with 'VALID' padding,
  the output would have the same dimensions as if the unpadded input was used
  in a convolution with 'SAME' padding.

  Args:
    inputs: A tensor of size [batch, height_in, width_in, channels].
    kernel_size: The kernel to be used in the conv2d or max_pool2d operation.
    rate: An integer, rate for atrous convolution.

  Returns:
    output: A tensor of size [batch, height_out, width_out, channels] with the
      input, either intact (if kernel_size == 1) or padded (if kernel_size > 1).
  """
  kernel_size_effective = [kernel_size[0] + (kernel_size[0] - 1) * (rate - 1),
                           kernel_size[0] + (kernel_size[0] - 1) * (rate - 1)]
  pad_total = [kernel_size_effective[0] - 1, kernel_size_effective[1] - 1]
  pad_beg = [pad_total[0] // 2, pad_total[1] // 2]
  pad_end = [pad_total[0] - pad_beg[0], pad_total[1] - pad_beg[1]]
  padded_inputs = tf.pad(
      tensor=inputs,
      paddings=[[0, 0], [pad_beg[0], pad_end[0]], [pad_beg[1], pad_end[1]],
                [0, 0]])
  return padded_inputs


def _make_divisible(v, divisor, min_value=None):
  if min_value is None:
    min_value = divisor
  new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
  # Make sure that round down does not go down by more than 10%.
  if new_v < 0.9 * v:
    new_v += divisor
  return int(new_v)


@contextlib.contextmanager
def _set_arg_scope_defaults(defaults):
  """Sets arg scope defaults for all items present in defaults.

  Args:
    defaults: dictionary/list of pairs, containing a mapping from
    function to a dictionary of default args.

  Yields:
    context manager where all defaults are set.
  """
  if hasattr(defaults, 'items'):
    items = list(defaults.items())
  else:
    items = defaults
  if not items:
    yield
  else:
    func, default_arg = items[0]
    with slim.arg_scope(func, **default_arg):
      with _set_arg_scope_defaults(items[1:]):
        yield


@slim.add_arg_scope
def depth_multiplier(output_params,
                     multiplier,
                     divisible_by=8,
                     min_depth=8,
                     **unused_kwargs):
  if 'num_outputs' not in output_params:
    return
  d = output_params['num_outputs']
  output_params['num_outputs'] = _make_divisible(d * multiplier, divisible_by,
                                                 min_depth)


_Op = collections.namedtuple('Op', ['op', 'params', 'multiplier_func'])


def op(opfunc, multiplier_func=depth_multiplier, **params):
  multiplier = params.pop('multiplier_transform', multiplier_func)
  return _Op(opfunc, params=params, multiplier_func=multiplier)


class NoOpScope(object):
  """No-op context manager."""

  def __enter__(self):
    return None

  def __exit__(self, exc_type, exc_value, traceback):
    return False


def safe_arg_scope(funcs, **kwargs):
  """Returns `slim.arg_scope` with all None arguments removed.

  Args:
    funcs: Functions to pass to `arg_scope`.
    **kwargs: Arguments to pass to `arg_scope`.

  Returns:
    arg_scope or No-op context manager.

  Note: can be useful if None value should be interpreted as "do not overwrite
    this parameter value".
  """
  filtered_args = {name: value for name, value in kwargs.items()
                   if value is not None}
  if filtered_args:
    return slim.arg_scope(funcs, **filtered_args)
  else:
    return NoOpScope()


@slim.add_arg_scope
def mobilenet_base(  # pylint: disable=invalid-name
    inputs,
    conv_defs,
    multiplier=1.0,
    final_endpoint=None,
    output_stride=None,
    use_explicit_padding=False,
    scope=None,
    is_training=False):
  """Mobilenet base network.

  Constructs a network from inputs to the given final endpoint. By default
  the network is constructed in inference mode. To create network
  in training mode use:

  with slim.arg_scope(mobilenet.training_scope()):
     logits, endpoints = mobilenet_base(...)

  Args:
    inputs: a tensor of shape [batch_size, height, width, channels].
    conv_defs: A list of op(...) layers specifying the net architecture.
    multiplier: Float multiplier for the depth (number of channels)
      for all convolution ops. The value must be greater than zero. Typical
      usage will be to set this value in (0, 1) to reduce the number of
      parameters or computation cost of the model.
    final_endpoint: The name of last layer, for early termination for
    for V1-based networks: last layer is "layer_14", for V2: "layer_20"
    output_stride: An integer that specifies the requested ratio of input to
      output spatial resolution. If not None, then we invoke atrous convolution
      if necessary to prevent the network from reducing the spatial resolution
      of the activation maps. Allowed values are 1 or any even number, excluding
      zero. Typical values are 8 (accurate fully convolutional mode), 16
      (fast fully convolutional mode), and 32 (classification mode).

      NOTE- output_stride relies on all consequent operators to support dilated
      operators via "rate" parameter. This might require wrapping non-conv
      operators to operate properly.

    use_explicit_padding: Use 'VALID' padding for convolutions, but prepad
      inputs so that the output dimensions are the same as if 'SAME' padding
      were used.
    scope: optional variable scope.
    is_training: How to setup batch_norm and other ops. Note: most of the time
      this does not need be set directly. Use mobilenet.training_scope() to set
      up training instead. This parameter is here for backward compatibility
      only. It is safe to set it to the value matching
      training_scope(is_training=...). It is also safe to explicitly set
      it to False, even if there is outer training_scope set to to training.
      (The network will be built in inference mode). If this is set to None,
      no arg_scope is added for slim.batch_norm's is_training parameter.

  Returns:
    tensor_out: output tensor.
    end_points: a set of activations for external use, for example summaries or
                losses.

  Raises:
    ValueError: depth_multiplier <= 0, or the target output_stride is not
                allowed.
  """
  if multiplier <= 0:
    raise ValueError('multiplier is not greater than zero.')

  # Set conv defs defaults and overrides.
  conv_defs_defaults = conv_defs.get('defaults', {})
  conv_defs_overrides = conv_defs.get('overrides', {})
  if use_explicit_padding:
    conv_defs_overrides = copy.deepcopy(conv_defs_overrides)
    conv_defs_overrides[
        (slim.conv2d, slim.separable_conv2d)] = {'padding': 'VALID'}

  if output_stride is not None:
    if output_stride == 0 or (output_stride > 1 and output_stride % 2):
      raise ValueError('Output stride must be None, 1 or a multiple of 2.')

  # a) Set the tensorflow scope
  # b) set padding to default: note we might consider removing this
  # since it is also set by mobilenet_scope
  # c) set all defaults
  # d) set all extra overrides.
  # pylint: disable=g-backslash-continuation
  with _scope_all(scope, default_scope='Mobilenet'), \
      safe_arg_scope([slim.batch_norm], is_training=is_training), \
      _set_arg_scope_defaults(conv_defs_defaults), \
      _set_arg_scope_defaults(conv_defs_overrides):
    # The current_stride variable keeps track of the output stride of the
    # activations, i.e., the running product of convolution strides up to the
    # current network layer. This allows us to invoke atrous convolution
    # whenever applying the next convolution would result in the activations
    # having output stride larger than the target output_stride.
    current_stride = 1

    # The atrous convolution rate parameter.
    rate = 1

    net = inputs
    # Insert default parameters before the base scope which includes
    # any custom overrides set in mobilenet.
    end_points = {}
    scopes = {}
    for i, opdef in enumerate(conv_defs['spec']):
      params = dict(opdef.params)
      opdef.multiplier_func(params, multiplier)
      stride = params.get('stride', 1)
      if output_stride is not None and current_stride == output_stride:
        # If we have reached the target output_stride, then we need to employ
        # atrous convolution with stride=1 and multiply the atrous rate by the
        # current unit's stride for use in subsequent layers.
        layer_stride = 1
        layer_rate = rate
        rate *= stride
      else:
        layer_stride = stride
        layer_rate = 1
        current_stride *= stride
      # Update params.
      params['stride'] = layer_stride
      # Only insert rate to params if rate > 1 and kernel size is not [1, 1].
      if layer_rate > 1:
        if tuple(params.get('kernel_size', [])) != (1, 1):
          # We will apply atrous rate in the following cases:
          # 1) When kernel_size is not in params, the operation then uses
          #   default kernel size 3x3.
          # 2) When kernel_size is in params, and if the kernel_size is not
          #   equal to (1, 1) (there is no need to apply atrous convolution to
          #   any 1x1 convolution).
          params['rate'] = layer_rate
      # Set padding
      if use_explicit_padding:
        if 'kernel_size' in params:
          net = _fixed_padding(net, params['kernel_size'], layer_rate)
        else:
          params['use_explicit_padding'] = True

      end_point = 'layer_%d' % (i + 1)
      try:
        net = opdef.op(net, **params)
      except Exception:
        print('Failed to create op %i: %r params: %r' % (i, opdef, params))
        raise
      end_points[end_point] = net
      scope = os.path.dirname(net.name)
      scopes[scope] = end_point
      if final_endpoint is not None and end_point == final_endpoint:
        break

    # Add all tensors that end with 'output' to
    # endpoints
    for t in net.graph.get_operations():
      scope = os.path.dirname(t.name)
      bn = os.path.basename(t.name)
      if scope in scopes and t.name.endswith('output'):
        end_points[scopes[scope] + '/' + bn] = t.outputs[0]
    return net, end_points


@contextlib.contextmanager
def _scope_all(scope, default_scope=None):
  with tf.variable_scope(scope, default_name=default_scope) as s,\
       tf.name_scope(s.original_name_scope):
    yield s


@slim.add_arg_scope
def mobilenet(inputs,
              num_classes=1001,
              prediction_fn=slim.softmax,
              reuse=None,
              scope='Mobilenet',
              base_only=False,
              use_reduce_mean_for_pooling=False,
              **mobilenet_args):
  """Mobilenet model for classification, supports both V1 and V2.

  Note: default mode is inference, use mobilenet.training_scope to create
  training network.


  Args:
    inputs: a tensor of shape [batch_size, height, width, channels].
    num_classes: number of predicted classes. If 0 or None, the logits layer
      is omitted and the input features to the logits layer (before dropout)
      are returned instead.
    prediction_fn: a function to get predictions out of logits
      (default softmax).
    reuse: whether or not the network and its variables should be reused. To be
      able to reuse 'scope' must be given.
    scope: Optional variable_scope.
    base_only: if True will only create the base of the network (no pooling
    and no logits).
    use_reduce_mean_for_pooling: if True use the reduce_mean for pooling. If
    True use the global_pool function that provides some optimization.
    **mobilenet_args: passed to mobilenet_base verbatim.
      - conv_defs: list of conv defs
      - multiplier: Float multiplier for the depth (number of channels)
      for all convolution ops. The value must be greater than zero. Typical
      usage will be to set this value in (0, 1) to reduce the number of
      parameters or computation cost of the model.
      - output_stride: will ensure that the last layer has at most total stride.
      If the architecture calls for more stride than that provided
      (e.g. output_stride=16, but the architecture has 5 stride=2 operators),
      it will replace output_stride with fractional convolutions using Atrous
      Convolutions.

  Returns:
    logits: the pre-softmax activations, a tensor of size
      [batch_size, num_classes]
    end_points: a dictionary from components of the network to the corresponding
      activation tensor.

  Raises:
    ValueError: Input rank is invalid.
  """
  is_training = mobilenet_args.get('is_training', False)
  input_shape = inputs.get_shape().as_list()
  if len(input_shape) != 4:
    raise ValueError('Expected rank 4 input, was: %d' % len(input_shape))

  with tf.variable_scope(scope, 'Mobilenet', reuse=reuse) as scope:
    inputs = tf.identity(inputs, 'input')
    net, end_points = mobilenet_base(inputs, scope=scope, **mobilenet_args)
    if base_only:
      return net, end_points

    net = tf.identity(net, name='embedding')

    with tf.variable_scope('Logits'):
      net = global_pool(net, use_reduce_mean_for_pooling)
      end_points['global_pool'] = net
      if not num_classes:
        return net, end_points
      net = slim.dropout(net, scope='Dropout', is_training=is_training)
      # 1 x 1 x num_classes
      # Note: legacy scope name.
      logits = slim.conv2d(
          net,
          num_classes, [1, 1],
          activation_fn=None,
          normalizer_fn=None,
          biases_initializer=tf.zeros_initializer(),
          scope='Conv2d_1c_1x1')

      logits = tf.squeeze(logits, [1, 2])

      logits = tf.identity(logits, name='output')
    end_points['Logits'] = logits
    if prediction_fn:
      end_points['Predictions'] = prediction_fn(logits, 'Predictions')
  return logits, end_points


def global_pool(input_tensor,
                use_reduce_mean_for_pooling=False,
                pool_op=tf.nn.avg_pool2d):
  """Applies avg pool to produce 1x1 output.

  NOTE: This function is funcitonally equivalenet to reduce_mean, but it has
  baked in average pool which has better support across hardware.

  Args:
    input_tensor: input tensor
    use_reduce_mean_for_pooling: if True use reduce_mean for pooling
    pool_op: pooling op (avg pool is default)
  Returns:
    a tensor batch_size x 1 x 1 x depth.
  """
  if use_reduce_mean_for_pooling:
    return tf.reduce_mean(
        input_tensor, [1, 2], keepdims=True, name='ReduceMean')
  else:
    shape = input_tensor.get_shape().as_list()
    if shape[1] is None or shape[2] is None:
      kernel_size = tf.convert_to_tensor(value=[
          1,
          tf.shape(input=input_tensor)[1],
          tf.shape(input=input_tensor)[2], 1
      ])
    else:
      kernel_size = [1, shape[1], shape[2], 1]
    output = pool_op(
        input_tensor, ksize=kernel_size, strides=[1, 1, 1, 1], padding='VALID')
    # Recover output shape, for unknown shape.
    output.set_shape([None, 1, 1, None])
    return output


def training_scope(is_training=True,
                   weight_decay=0.00004,
                   stddev=0.09,
                   dropout_keep_prob=0.8,
                   bn_decay=0.997):
  """Defines Mobilenet training scope.

  Usage:
     with slim.arg_scope(mobilenet.training_scope()):
       logits, endpoints = mobilenet_v2.mobilenet(input_tensor)

     # the network created will be trainble with dropout/batch norm
     # initialized appropriately.
  Args:
    is_training: if set to False this will ensure that all customizations are
      set to non-training mode. This might be helpful for code that is reused
      across both training/evaluation, but most of the time training_scope with
      value False is not needed. If this is set to None, the parameters is not
      added to the batch_norm arg_scope.

    weight_decay: The weight decay to use for regularizing the model.
    stddev: Standard deviation for initialization, if negative uses xavier.
    dropout_keep_prob: dropout keep probability (not set if equals to None).
    bn_decay: decay for the batch norm moving averages (not set if equals to
      None).

  Returns:
    An argument scope to use via arg_scope.
  """
  # Note: do not introduce parameters that would change the inference
  # model here (for example whether to use bias), modify conv_def instead.
  batch_norm_params = {
      'decay': bn_decay,
      'is_training': is_training
  }
  if stddev < 0:
    weight_intitializer = slim.initializers.xavier_initializer()
  else:
    weight_intitializer = tf.truncated_normal_initializer(
        stddev=stddev)

  # Set weight_decay for weights in Conv and FC layers.
  with slim.arg_scope(
      [slim.conv2d, slim.fully_connected, slim.separable_conv2d],
      weights_initializer=weight_intitializer,
      normalizer_fn=slim.batch_norm), \
      slim.arg_scope([mobilenet_base, mobilenet], is_training=is_training),\
      safe_arg_scope([slim.batch_norm], **batch_norm_params), \
      safe_arg_scope([slim.dropout], is_training=is_training,
                     keep_prob=dropout_keep_prob), \
      slim.arg_scope([slim.conv2d], \
                     weights_regularizer=slim.l2_regularizer(weight_decay)), \
      slim.arg_scope([slim.separable_conv2d], weights_regularizer=None) as s:
    return s



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for utils.target_assigner_utils."""

from absl.testing import parameterized
import numpy as np
import tensorflow.compat.v1 as tf

from ..core import box_list
from ..utils import target_assigner_utils as ta_utils
from ..utils import test_case


class TargetUtilTest(parameterized.TestCase, test_case.TestCase):

  def test_image_shape_to_grids(self):
    def graph_fn():
      (y_grid, x_grid) = ta_utils.image_shape_to_grids(height=2, width=3)
      return y_grid, x_grid

    expected_y_grid = np.array([[0, 0, 0], [1, 1, 1]])
    expected_x_grid = np.array([[0, 1, 2], [0, 1, 2]])

    y_grid, x_grid = self.execute(graph_fn, [])

    np.testing.assert_array_equal(y_grid, expected_y_grid)
    np.testing.assert_array_equal(x_grid, expected_x_grid)

  @parameterized.parameters((False,), (True,))
  def test_coordinates_to_heatmap(self, sparse):
    if not hasattr(tf, 'tensor_scatter_nd_max'):
      self.skipTest('Cannot test function due to old TF version.')

    def graph_fn():
      (y_grid, x_grid) = ta_utils.image_shape_to_grids(height=3, width=5)
      y_coordinates = tf.constant([1.5, 0.5], dtype=tf.float32)
      x_coordinates = tf.constant([2.5, 4.5], dtype=tf.float32)
      sigma = tf.constant([0.1, 0.5], dtype=tf.float32)
      channel_onehot = tf.constant([[1, 0, 0], [0, 1, 0]], dtype=tf.float32)
      channel_weights = tf.constant([1, 1], dtype=tf.float32)
      heatmap = ta_utils.coordinates_to_heatmap(y_grid, x_grid, y_coordinates,
                                                x_coordinates, sigma,
                                                channel_onehot,
                                                channel_weights, sparse=sparse)
      return heatmap

    heatmap = self.execute(graph_fn, [])
    # Peak at (1, 2) for the first class.
    self.assertAlmostEqual(1.0, heatmap[1, 2, 0])
    # Peak at (0, 4) for the second class.
    self.assertAlmostEqual(1.0, heatmap[0, 4, 1])

  def test_compute_floor_offsets_with_indices_onlysource(self):

    def graph_fn():
      y_source = tf.constant([1.5, 0.3], dtype=tf.float32)
      x_source = tf.constant([2.5, 4.2], dtype=tf.float32)
      (offsets, indices) = ta_utils.compute_floor_offsets_with_indices(
          y_source, x_source)

      return offsets, indices

    offsets, indices = self.execute(graph_fn, [])

    np.testing.assert_array_almost_equal(offsets,
                                         np.array([[0.5, 0.5], [0.3, 0.2]]))
    np.testing.assert_array_almost_equal(indices,
                                         np.array([[1, 2], [0, 4]]))

  def test_compute_floor_offsets_with_indices_and_targets(self):

    def graph_fn():
      y_source = tf.constant([1.5, 0.3], dtype=tf.float32)
      x_source = tf.constant([2.5, 4.2], dtype=tf.float32)
      y_target = tf.constant([2.1, 0.1], dtype=tf.float32)
      x_target = tf.constant([1.2, 4.5], dtype=tf.float32)
      (offsets, indices) = ta_utils.compute_floor_offsets_with_indices(
          y_source, x_source, y_target, x_target)
      return offsets, indices

    offsets, indices = self.execute(graph_fn, [])

    np.testing.assert_array_almost_equal(offsets,
                                         np.array([[1.1, -0.8], [0.1, 0.5]]))
    np.testing.assert_array_almost_equal(indices, np.array([[1, 2], [0, 4]]))

  def test_compute_floor_offsets_with_indices_multisources(self):

    def graph_fn():
      y_source = tf.constant([[1.0, 0.0], [2.0, 3.0]], dtype=tf.float32)
      x_source = tf.constant([[2.0, 4.0], [3.0, 3.0]], dtype=tf.float32)
      y_target = tf.constant([2.1, 0.1], dtype=tf.float32)
      x_target = tf.constant([1.2, 4.5], dtype=tf.float32)
      (offsets, indices) = ta_utils.compute_floor_offsets_with_indices(
          y_source, x_source, y_target, x_target)
      return offsets, indices

    offsets, indices = self.execute(graph_fn, [])
    # Offset from the first source to target.
    np.testing.assert_array_almost_equal(offsets[:, 0, :],
                                         np.array([[1.1, -0.8], [-1.9, 1.5]]))
    # Offset from the second source to target.
    np.testing.assert_array_almost_equal(offsets[:, 1, :],
                                         np.array([[2.1, -2.8], [-2.9, 1.5]]))
    # Indices from the first source to target.
    np.testing.assert_array_almost_equal(indices[:, 0, :],
                                         np.array([[1, 2], [2, 3]]))
    # Indices from the second source to target.
    np.testing.assert_array_almost_equal(indices[:, 1, :],
                                         np.array([[0, 4], [3, 3]]))

  def test_get_valid_keypoints_mask(self):

    def graph_fn():
      class_onehot = tf.constant(
          [[0, 0, 1, 0, 0],
           [0, 1, 0, 0, 0],
           [0, 0, 1, 0, 1]], dtype=tf.float32)
      keypoints = tf.constant(
          [[0.1, float('nan'), 0.2, 0.0],
           [0.0, 0.0, 0.1, 0.9],
           [3.2, 4.3, float('nan'), 0.2]],
          dtype=tf.float32)
      keypoint_coordinates = tf.stack([keypoints, keypoints], axis=2)
      mask, keypoints_nan_to_zeros = ta_utils.get_valid_keypoint_mask_for_class(
          keypoint_coordinates=keypoint_coordinates,
          class_id=2,
          class_onehot=class_onehot,
          keypoint_indices=[1, 2])

      return mask, keypoints_nan_to_zeros

    keypoints = np.array([[0.0, 0.2],
                          [0.0, 0.1],
                          [4.3, 0.0]])
    expected_mask = np.array([[0, 1], [0, 0], [1, 0]])
    expected_keypoints = np.stack([keypoints, keypoints], axis=2)
    mask, keypoints_nan_to_zeros = self.execute(graph_fn, [])

    np.testing.assert_array_equal(mask, expected_mask)
    np.testing.assert_array_almost_equal(keypoints_nan_to_zeros,
                                         expected_keypoints)

  def test_get_valid_keypoints_with_mask(self):
    def graph_fn():
      class_onehot = tf.constant(
          [[0, 0, 1, 0, 0],
           [0, 1, 0, 0, 0],
           [0, 0, 1, 0, 1]], dtype=tf.float32)
      keypoints = tf.constant(
          [[0.1, float('nan'), 0.2, 0.0],
           [0.0, 0.0, 0.1, 0.9],
           [3.2, 4.3, float('nan'), 0.2]],
          dtype=tf.float32)
      keypoint_coordinates = tf.stack([keypoints, keypoints], axis=2)
      weights = tf.constant([0.0, 0.0, 1.0])
      mask, keypoints_nan_to_zeros = ta_utils.get_valid_keypoint_mask_for_class(
          keypoint_coordinates=keypoint_coordinates,
          class_id=2,
          class_onehot=class_onehot,
          class_weights=weights,
          keypoint_indices=[1, 2])
      return mask, keypoints_nan_to_zeros

    expected_mask = np.array([[0, 0], [0, 0], [1, 0]])
    keypoints = np.array([[0.0, 0.2],
                          [0.0, 0.1],
                          [4.3, 0.0]])
    expected_keypoints = np.stack([keypoints, keypoints], axis=2)
    mask, keypoints_nan_to_zeros = self.execute(graph_fn, [])

    np.testing.assert_array_equal(mask, expected_mask)
    np.testing.assert_array_almost_equal(keypoints_nan_to_zeros,
                                         expected_keypoints)

  def test_blackout_pixel_weights_by_box_regions(self):
    def graph_fn():
      boxes = tf.constant(
          [[0.0, 0.0, 5, 5], [0.0, 0.0, 10.0, 20.0], [6.0, 12.0, 8.0, 18.0]],
          dtype=tf.float32)
      blackout = tf.constant([True, False, True], dtype=tf.bool)
      blackout_pixel_weights_by_box_regions = tf.function(
          ta_utils.blackout_pixel_weights_by_box_regions)
      output = blackout_pixel_weights_by_box_regions(10, 20, boxes, blackout)
      return output

    output = self.execute(graph_fn, [])
    # All zeros in region [0:5, 0:5].
    self.assertAlmostEqual(np.sum(output[0:5, 0:5]), 0.0)
    # All zeros in region [12:18, 6:8].
    self.assertAlmostEqual(np.sum(output[6:8, 12:18]), 0.0)
    # All other pixel weights should be 1.0.
    # 20 * 10 - 5 * 5 - 2 * 6 = 163.0
    self.assertAlmostEqual(np.sum(output), 163.0)

  def test_blackout_pixel_weights_by_box_regions_with_weights(self):
    def graph_fn():
      boxes = tf.constant(
          [[0.0, 0.0, 2.0, 2.0],
           [0.0, 0.0, 4.0, 2.0],
           [3.0, 0.0, 4.0, 4.0]],
          dtype=tf.float32)
      blackout = tf.constant([False, False, True], dtype=tf.bool)
      weights = tf.constant([0.4, 0.3, 0.2], tf.float32)
      blackout_pixel_weights_by_box_regions = tf.function(
          ta_utils.blackout_pixel_weights_by_box_regions)
      output = blackout_pixel_weights_by_box_regions(
          4, 4, boxes, blackout, weights)
      return output

    output = self.execute(graph_fn, [])
    expected_weights = [
        [0.4, 0.4, 1.0, 1.0],
        [0.4, 0.4, 1.0, 1.0],
        [0.3, 0.3, 1.0, 1.0],
        [0.0, 0.0, 0.0, 0.0]]
    np.testing.assert_array_almost_equal(expected_weights, output)

  def test_blackout_pixel_weights_by_box_regions_zero_instance(self):
    def graph_fn():
      boxes = tf.zeros([0, 4], dtype=tf.float32)
      blackout = tf.zeros([0], dtype=tf.bool)
      blackout_pixel_weights_by_box_regions = tf.function(
          ta_utils.blackout_pixel_weights_by_box_regions)
      output = blackout_pixel_weights_by_box_regions(10, 20, boxes, blackout)
      return output

    output = self.execute(graph_fn, [])
    # The output should be all 1s since there's no annotation provided.
    np.testing.assert_array_equal(output, np.ones([10, 20], dtype=np.float32))

  def test_get_surrounding_grids(self):

    def graph_fn():
      y_coordinates = tf.constant([0.5], dtype=tf.float32)
      x_coordinates = tf.constant([4.5], dtype=tf.float32)
      output = ta_utils.get_surrounding_grids(
          height=3,
          width=5,
          y_coordinates=y_coordinates,
          x_coordinates=x_coordinates,
          radius=1)
      return output

    y_indices, x_indices, valid = self.execute(graph_fn, [])

    # Five neighboring indices: [-1, 4] (out of bound), [0, 3], [0, 4],
    # [0, 5] (out of bound), [1, 4].
    np.testing.assert_array_almost_equal(
        y_indices,
        np.array([[0.0, 0.0, 0.0, 0.0, 1.0]]))
    np.testing.assert_array_almost_equal(
        x_indices,
        np.array([[0.0, 3.0, 4.0, 0.0, 4.0]]))
    self.assertAllEqual(valid, [[False, True, True, False, True]])

  def test_coordinates_to_iou(self):

    def graph_fn():
      y, x = tf.meshgrid(tf.range(32, dtype=tf.float32),
                         tf.range(32, dtype=tf.float32))
      blist = box_list.BoxList(
          tf.constant([[0., 0., 32., 32.],
                       [0., 0., 16., 16.],
                       [0.0, 0.0, 4.0, 4.0]]))
      classes = tf.constant([[0., 1., 0.],
                             [1., 0., 0.],
                             [0., 0., 1.]])

      result = ta_utils.coordinates_to_iou(
          y, x, blist, classes)
      return result

    result = self.execute(graph_fn, [])
    self.assertEqual(result.shape, (32, 32, 3))
    self.assertAlmostEqual(result[0, 0, 0], 1.0 / 7.0)
    self.assertAlmostEqual(result[0, 0, 1], 1.0 / 7.0)
    self.assertAlmostEqual(result[0, 16, 0], 1.0 / 7.0)
    self.assertAlmostEqual(result[2, 2, 2], 1.0)
    self.assertAlmostEqual(result[8, 8, 2], 0.0)


if __name__ == '__main__':
  tf.test.main()



from unittest import skip
import cv2
import numpy as np

def get_data(input_path, cat = None, skip_header = False, path_prefix = ""):

	found_bg = False
	all_imgs = {}

	classes_count = {}

	class_mapping = {}

	visualise = True
	
	with open(input_path,'r') as f:

		print('Parsing annotation files')

		for line in f:
			if skip_header:
				skip_header = False
				continue # skip 1st row

			line_split = line.strip().split(',')
			(filename,x1,y1,x2,y2,class_name) = line_split

			if class_name not in classes_count:
				classes_count[class_name] = 1
			else:
				classes_count[class_name] += 1

			if class_name not in class_mapping:
				if class_name == 'bg' and found_bg == False:
					print('Found class name with special name bg. Will be treated as a background region (this is usually for hard negative mining).')
					found_bg = True
				class_mapping[class_name] = len(class_mapping)

			if filename not in all_imgs:
				all_imgs[filename] = {}
				print(path_prefix + filename)
				img = cv2.imread(path_prefix + filename)

				(rows,cols) = img.shape[:2]
				# print(rows, cols)

				all_imgs[filename]['filepath'] = filename
				all_imgs[filename]['width'] = cols
				all_imgs[filename]['height'] = rows
				all_imgs[filename]['bboxes'] = []
				if np.random.randint(0,6) > 0:
					all_imgs[filename]['imageset'] = 'trainval'
				else:
					all_imgs[filename]['imageset'] = 'test'

			all_imgs[filename]['bboxes'].append({'class': class_name, 'x1': round(float(x1)), 'x2': round(float(x2)), 'y1': round(float(y1)), 'y2': round(float(y2))})


		all_data = []
		for key in all_imgs:
			all_data.append(all_imgs[key])
		
		# make sure the bg class is last in the list
		if found_bg:
			if class_mapping['bg'] != len(class_mapping) - 1:
				key_to_switch = [key for key in class_mapping.keys() if class_mapping[key] == len(class_mapping)-1][0]
				val_to_switch = class_mapping['bg']
				class_mapping['bg'] = len(class_mapping) - 1
				class_mapping[key_to_switch] = val_to_switch
		
		return all_data, classes_count, class_mapping


